import pytest
import requests
from requests.exceptions import (
    ConnectTimeout,
    ConnectionError,
    RequestException,
    Timeout,
)
from requests.models import Response

from isar.services.service_connections.request_handler import RequestHandler

url = "http://10.0.0.1"


def test_request_handler_success(requests_mock):
    base_handler = RequestHandler()
    requests_mock.get(url, json=[{"Success": 1}], status_code=200)
    requests_mock.delete(url, json=[{"Success": 1}], status_code=200)
    requests_mock.put(url, json=[{"Success": 1}], status_code=200)
    requests_mock.post(url, json=[{"Success": 1}], status_code=200)

    put_response = base_handler.put(url)
    get_response = base_handler.get(url)
    post_response = base_handler.post(url)
    delete_response = base_handler.delete(url)

    assert isinstance(put_response, Response)
    assert put_response.status_code == 200

    assert isinstance(get_response, Response)
    assert get_response.status_code == 200

    assert isinstance(post_response, Response)
    assert post_response.status_code == 200

    assert isinstance(delete_response, Response)
    assert delete_response.status_code == 200


def test_request_http_error(requests_mock):
    base_handler = RequestHandler()
    requests_mock.get(url, json=[], status_code=400)
    requests_mock.delete(url, json=[], status_code=400)
    requests_mock.put(url, json=[], status_code=400)
    requests_mock.post(url, json=[{"Success": 1}], status_code=400)
    with pytest.raises(RequestException):
        base_handler.post(url)
    with pytest.raises(RequestException):
        base_handler.put(url)
    with pytest.raises(RequestException):
        base_handler.delete(url)
    with pytest.raises(RequestException):
        base_handler.get(url)


@pytest.mark.parametrize(
    "exception",
    [Timeout(), ConnectionError(), KeyError(), ConnectTimeout()],
)
def test_request_exception(mocker, exception):
    base_handler = RequestHandler()
    mocker.patch.object(requests, "request", side_effect=exception)
    with pytest.raises(RequestException):
        base_handler.put(url)
    with pytest.raises(RequestException):
        base_handler.post(url)
    with pytest.raises(RequestException):
        base_handler.delete(url)
    with pytest.raises(RequestException):
        base_handler.get(url)


def test_timeout_exception():
    base_handler = RequestHandler()
    with pytest.raises(RequestException):
        base_handler.put(url, request_timeout=1e-10)
    with pytest.raises(RequestException):
        base_handler.post(url, request_timeout=1e-10)
    with pytest.raises(RequestException):
        base_handler.delete(url, request_timeout=1e-10)
    with pytest.raises(RequestException):
        base_handler.get(url, request_timeout=1e-10)



import logging
from logging import Logger
from threading import Thread
from typing import List

from dotenv import load_dotenv
from injector import Injector

from isar.apis.api import API
from isar.config.log import setup_logger
from isar.config.settings import settings
from isar.models.communication.queues.queues import Queues
from isar.modules import get_injector
from isar.services.service_connections.mqtt.mqtt_client import MqttClient
from isar.state_machine.state_machine import main
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.robot_interface import RobotInterface

if __name__ == "__main__":
    load_dotenv()

    setup_logger()
    logger: Logger = logging.getLogger("main")

    injector: Injector = get_injector()

    robot: RobotInterface = injector.get(RobotInterface)
    queues: Queues = injector.get(Queues)

    threads: List[Thread] = []

    state_machine_thread: Thread = Thread(
        target=main, name="ISAR State Machine", args=[injector], daemon=True
    )
    threads.append(state_machine_thread)

    uploader: Uploader = Uploader(
        upload_queue=queues.upload_queue,
        storage_handlers=injector.get(List[StorageInterface]),
    )

    uploader_thread: Thread = Thread(
        target=uploader.run, name="ISAR Uploader", daemon=True
    )
    threads.append(uploader_thread)

    if settings.MQTT_ENABLED:
        mqtt_client: MqttClient = MqttClient(mqtt_queue=queues.mqtt_queue)

        mqtt_thread: Thread = Thread(
            target=mqtt_client.run, name="ISAR MQTT Client", daemon=True
        )

        threads.append(mqtt_thread)

        publishers: List[Thread] = robot.get_telemetry_publishers(
            queue=queues.mqtt_queue, robot_id=settings.ROBOT_ID
        )
        threads.extend(publishers)

    api: API = injector.get(API)
    api_thread: Thread = Thread(target=api.run_app, name="ISAR API", daemon=True)
    threads.append(api_thread)

    for thread in threads:
        thread.start()
        logger.info(f"Started thread: {thread.name}")

    for thread in threads:
        thread.join()
        logger.info(f"Joined thread: {thread.name}")



import importlib.resources as pkg_resources
import logging
import logging.config

import yaml
from uvicorn.logging import ColourizedFormatter

from isar.config.settings import settings


def setup_logger():
    log_levels: dict = settings.LOG_LEVELS
    with pkg_resources.path("isar.config", "logging.conf") as path:
        log_config = yaml.safe_load(open(path))

    log_handler = logging.StreamHandler()

    log_handler.setLevel(log_config["root"]["level"])
    log_handler.setFormatter(
        ColourizedFormatter(
            log_config["formatters"]["colourized"]["format"],
            style="{",
            use_colors=True,
        )
    )

    logging.config.dictConfig(log_config)

    for loggers in log_config["loggers"].keys():
        logging.getLogger(loggers).addHandler(log_handler)
        logging.getLogger(loggers).setLevel(log_levels[loggers])
    logging.getLogger().addHandler(log_handler)



from typing import List, Optional

from alitra import Position
from pydantic import BaseModel, Field

from isar.apis.models.models import InputPose, InputPosition
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.models.mission.mission import Mission, Task
from robot_interface.models.mission.step import (
    DriveToPose,
    InspectionStep,
    STEPS,
    TakeImage,
    TakeThermalImage,
    TakeThermalVideo,
    TakeVideo,
)

inspection_step_types: List[type[InspectionStep]] = [
    TakeImage,
    TakeThermalImage,
    TakeVideo,
    TakeThermalVideo,
]


class StartMissionTaskDefinition(BaseModel):
    pose: InputPose
    tag: Optional[str]
    inspection_target: InputPosition
    inspection_types: List[str] = Field(
        default=[step.get_inspection_type().__name__ for step in inspection_step_types]
    )
    video_duration: Optional[float] = Field(default=10)


class StartMissionDefinition(BaseModel):
    tasks: List[StartMissionTaskDefinition]


def to_isar_mission(mission_definition: StartMissionDefinition) -> Mission:
    isar_tasks: List[Task] = []

    for task in mission_definition.tasks:
        try:
            tag_id: Optional[str] = task.tag
            drive_step: DriveToPose = DriveToPose(pose=task.pose.to_alitra_pose())
            inspection_steps: List[STEPS] = [
                create_inspection_step(
                    inspection_type=inspection_type,
                    duration=task.video_duration,
                    target=task.inspection_target.to_alitra_position(),
                )
                for inspection_type in task.inspection_types
            ]
        except (ValueError) as e:
            raise MissionPlannerError(f"Failed to create task: {str(e)}")
        isar_task: Task = Task(steps=[drive_step, *inspection_steps], tag_id=tag_id)
        isar_tasks.append(isar_task)

    if not isar_tasks:
        raise MissionPlannerError("Mission does not contain any valid tasks")

    isar_mission: Mission = Mission(tasks=isar_tasks)

    return isar_mission


def create_inspection_step(
    inspection_type: str, duration: float, target: Position
) -> STEPS:
    inspection: STEPS

    if inspection_type == TakeImage.get_inspection_type().__name__:
        inspection = TakeImage(target=target)
    elif inspection_type == TakeVideo.get_inspection_type().__name__:
        inspection = TakeVideo(target=target, duration=duration)
    elif inspection_type == TakeThermalImage.get_inspection_type().__name__:
        inspection = TakeThermalImage(target=target)
    elif inspection_type == TakeThermalVideo.get_inspection_type().__name__:
        inspection = TakeThermalVideo(target=target, duration=duration)
    else:
        raise ValueError(f"Inspection type '{inspection_type}' not supported")

    return inspection



import pytest

from isar.models.communication.queues import QueueIO, QueueTimeoutError
from isar.services.utilities.queue_utilities import QueueUtilities


def test_timeout_send_command(mocker, scheduling_utilities):
    mocker.patch.object(QueueUtilities, "check_queue", side_effect=QueueTimeoutError)
    q: QueueIO = QueueIO(input_size=1, output_size=1)
    with pytest.raises(QueueTimeoutError):
        scheduling_utilities._send_command(True, q)
    assert q.input.empty()


import queue
from queue import Queue



from .status import StepStatus
from .step import (
    DriveToPose,
    InspectionStep,
    MotionStep,
    STEPS,
    Step,
    TakeImage,
    TakeThermalImage,
    TakeThermalVideo,
    TakeVideo,
)



import logging
import time
from typing import TYPE_CHECKING, Callable

from transitions import State

from isar.config.settings import settings
from isar.models.mission.status import MissionStatus
from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions import (
    RobotException,
    RobotInfeasibleStepException,
)
from robot_interface.models.mission.status import StepStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class InitiateStep(State):
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="initiate_step", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.initiate_step_failure_counter: int = 0
        self.initiate_step_failure_counter_limit: int = (
            settings.INITIATE_STEP_FAILURE_COUNTER_LIMIT
        )
        self.logger = logging.getLogger("state_machine")

        self.initiate_step_thread = None

    def start(self):
        self.state_machine.update_state()
        self._run()

    def stop(self):
        self.initiate_step_failure_counter = 0
        if self.initiate_step_thread:
            self.initiate_step_thread.wait_for_thread()
        self.initiate_step_thread = None

    def _run(self):
        transition: Callable
        while True:
            if self.state_machine.should_stop_mission():
                transition = self.state_machine.stop
                break

            if self.state_machine.should_pause_mission():
                transition = self.state_machine.pause
                break

            if not self.state_machine.current_task:
                self.logger.info(
                    f"Completed mission: {self.state_machine.current_mission.id}"
                )
                transition = self.state_machine.mission_finished
                break

            if not self.initiate_step_thread:
                self.initiate_step_thread = ThreadedRequest(
                    self.state_machine.robot.initiate_step
                )
                self.initiate_step_thread.start_thread(self.state_machine.current_step)

            try:
                self.initiate_step_thread.get_output()
                transition = self.state_machine.step_initiated
                break
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue
            except RobotInfeasibleStepException:
                self.logger.warning(
                    f"Failed to initiate "
                    f"{type(self.state_machine.current_step).__name__}"
                    f"Invalid step: {str(self.state_machine.current_step.id)[:8]}"
                )
                transition = self.state_machine.step_infeasible
                break
            except RobotException as e:
                self.initiate_step_thread = None
                self.initiate_step_failure_counter += 1
                self.logger.warning(
                    f"Initiating step failed #: "
                    f"{str(self.initiate_step_failure_counter)}"
                    f"{e}"
                )

            if (
                self.initiate_step_failure_counter
                >= self.initiate_step_failure_counter_limit
            ):
                self.logger.error(
                    f"Failed to initiate step after "
                    f"{self.initiate_step_failure_counter_limit} attempts. "
                    f"Cancelling mission."
                )
                transition = self.state_machine.initiate_step_failed
                break

            time.sleep(self.state_machine.sleep_time)

        transition()



from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Any, List, Literal, Optional, Type, Union
from uuid import UUID, uuid4

from alitra import Pose, Position

from robot_interface.models.inspection.inspection import (
    Image,
    Inspection,
    ThermalImage,
    ThermalVideo,
    Video,
)
from robot_interface.models.mission.status import StepStatus


@dataclass
class Step:
    """
    Base class for all steps in a mission.
    """

    id: UUID = field(default_factory=uuid4, init=False)
    status: StepStatus = field(default=StepStatus.NotStarted, init=False)

    def __str__(self):
        def add_indent(text: str) -> str:
            return "".join("  " + line for line in text.splitlines(True))

        def robot_class_to_pretty_string(obj: Step) -> str:
            log_message: str = ""
            for attr in dir(obj):
                if callable(getattr(obj, attr)) or attr.startswith("__"):
                    continue

                value: Any = getattr(obj, attr)
                try:
                    package_name: Optional[str] = (
                        str(value.__class__).split("'")[1].split(".")[0]
                    )
                except (AttributeError, IndexError):
                    package_name = None

                if package_name == "robot_interface":
                    log_message += (
                        "\n" + attr + ": " + robot_class_to_pretty_string(value)
                    )
                else:
                    log_message += "\n" + str(attr) + ": " + str(value)

            return add_indent(log_message)

        class_name: str = type(self).__name__
        return class_name + robot_class_to_pretty_string(self)


@dataclass
class InspectionStep(Step):
    """
    Base class for all inspection steps which produce results to be uploaded.
    """

    inspections: List[Inspection] = field(default_factory=list, init=False)
    tag_id: Optional[str] = field(default=None, init=False)
    type = "inspection_type"

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return Inspection

    pass


@dataclass
class MotionStep(Step):
    """
    Base class for all steps which should move the robot, but not return a result.
    """

    pass


@dataclass
class ContinousInspectionStep(Step):
    """
    Base class for all continous inspection steps which produce a result to be uploaded.
    """

    pass


@dataclass
class DriveToPose(MotionStep):
    """
    Step which causes the robot to move to the given pose.
    """

    pose: Pose
    type: Literal["drive_to_pose"] = "drive_to_pose"


@dataclass
class DockingProcedure(MotionStep):
    """
    Step which causes the robot to dock or undock
    """

    behavior: Literal["dock, undock"]
    type: Literal["docking_procedure"] = "docking_procedure"


@dataclass
class TakeImage(InspectionStep):
    """
    Step which causes the robot to take an image towards the given coordinate.
    """

    target: Position
    type: Literal["take_image"] = "take_image"

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return Image


@dataclass
class TakeThermalImage(InspectionStep):
    """
    Step which causes the robot to take a thermal image towards the given coordinate.
    """

    target: Position
    type: Literal["take_thermal_image"] = "take_thermal_image"

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return ThermalImage


@dataclass
class TakeVideo(InspectionStep):
    """
    Step which causes the robot to take a video towards the given coordinate.

    Duration of video is given in seconds.
    """

    target: Position
    duration: float
    type: Literal["take_video"] = "take_video"

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return Video


@dataclass
class TakeThermalVideo(InspectionStep):
    """
    Step which causes the robot to record thermal video towards the given coordinate

    Duration of video is given in seconds.
    """

    target: Position
    duration: float
    type: Literal["take_thermal_video"] = "take_thermal_video"

    @staticmethod
    def get_inspection_type() -> Type[Inspection]:
        return ThermalVideo


STEPS = Union[
    DriveToPose,
    DockingProcedure,
    TakeImage,
    TakeThermalImage,
    TakeVideo,
    TakeThermalVideo,
]



from dataclasses import asdict
from typing import Any

import pytest
from alitra import Pose

from isar.models.mission import Mission
from isar.services.readers.base_reader import BaseReader
from robot_interface.models.mission import Step
from tests.mocks.mission_definition import MockMissionDefinition
from tests.mocks.pose import MockPose
from tests.mocks.step import MockStep


class TestBaseReader:
    @pytest.mark.parametrize(
        "dataclass_dict, expected_dataclass",
        [
            (asdict(MockMissionDefinition.default_mission), Mission),
            (asdict(MockStep.drive_to), Step),
            (asdict(MockStep.take_image_in_coordinate_direction), Step),
            (asdict(MockPose.default_pose), Pose),
        ],
    )
    def test_dict_to_dataclass(self, dataclass_dict: dict, expected_dataclass: Any):
        content = BaseReader.dict_to_dataclass(dataclass_dict, expected_dataclass)
        assert type(content) is expected_dataclass



import pytest

from isar.models.mission import Mission, Task
from isar.state_machine.states.monitor import Monitor
from robot_interface.models.mission import Step, StepStatus, TakeImage
from tests.mocks.step import MockStep


@pytest.mark.parametrize(
    "mock_status, expected_output",
    [
        (StepStatus.Successful, True),
        (StepStatus.Successful, True),
        (StepStatus.Failed, True),
    ],
)
def test_step_finished(monitor: Monitor, mock_status, expected_output):
    step: Step = MockStep.drive_to
    step.status = mock_status
    step_completed: bool = monitor._step_finished(
        step=step,
    )

    assert step_completed == expected_output


@pytest.mark.parametrize(
    "mock_status, should_queue_upload",
    [
        (StepStatus.Successful, True),
        (StepStatus.Failed, False),
    ],
)
def test_should_only_upload_if_status_is_completed(
    monitor: Monitor, mock_status, should_queue_upload
):
    step: TakeImage = MockStep.take_image_in_coordinate_direction
    step.status = mock_status
    task: Task = Task(steps=[step])
    mission: Mission = Mission(tasks=[task])

    monitor.state_machine.current_mission = mission
    monitor._process_finished_step(step)
    assert monitor.state_machine.queues.upload_queue.empty() == (
        not should_queue_upload
    )



from abc import ABCMeta, abstractmethod

from isar.models.mission_metadata.mission_metadata import MissionMetadata
from robot_interface.models.inspection.inspection import Inspection


class StorageInterface(metaclass=ABCMeta):
    @abstractmethod
    def store(self, inspection: Inspection, metadata: MissionMetadata):
        """
        Parameters
        ----------
        metadata : MissionMetadata
            Metadata for the mission the inspection is a part of.
        inspection : Inspection
            The inspection object to be stored.

        Raises
        ----------
        StorageException
            An error occurred when storing the inspection.
        """
        pass


class StorageException(Exception):
    pass



from dataclasses import dataclass
from datetime import date, datetime
from typing import Optional, Union
from uuid import UUID

from isar.config.settings import settings

additional_meta: dict = {}


@dataclass
class MissionMetadata:
    mission_id: Union[UUID, int, str, None]
    coordinate_reference_system: str = settings.COORDINATE_REFERENCE_SYSTEM
    vertical_reference_system: str = settings.VERTICAL_REFERENCE_SYSTEM
    data_classification: str = settings.DATA_CLASSIFICATION
    source_url: Optional[str] = None
    plant_code: str = settings.PLANT_CODE
    plant_name: str = settings.PLANT_NAME
    media_orientation_reference_system: str = (
        settings.MEDIA_ORIENTATION_REFERENCE_SYSTEM
    )
    robot_id: str = settings.ROBOT_ID
    mission_date: date = datetime.utcnow().date()



from enum import Enum


class States(str, Enum):

    Off = "off"
    Idle = "idle"
    InitiateStep = "initiate_step"
    Initialize = "initialize"
    Monitor = "monitor"
    Paused = "paused"
    StopStep = "stop_step"

    def __repr__(self):
        return self.value



import json
import logging
from pathlib import Path
from typing import Any, Tuple
from uuid import UUID

from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.storage.storage_interface import StorageException
from robot_interface.models.inspection import ThermalVideo, Video
from robot_interface.models.inspection.inspection import Image, Inspection, ThermalImage
from robot_interface.utilities.json_service import EnhancedJSONEncoder


def construct_local_paths(
    inspection: Inspection, metadata: MissionMetadata
) -> Tuple[Path, Path]:
    inspection_type: str = get_inspection_type(inspection=inspection)

    folder: Path = Path(str(metadata.mission_id))
    filename: str = get_filename(
        mission_id=metadata.mission_id,
        inspection_type=inspection_type,
        inspection_id=inspection.id,
    )

    inspection_path: Path = folder.joinpath(
        f"{filename}.{inspection.metadata.file_type}"
    )

    metadata_path: Path = folder.joinpath(f"{filename}.json")

    return inspection_path, metadata_path


def construct_metadata_file(
    inspection: Inspection, metadata: MissionMetadata, filename: str
) -> bytes:
    data: dict = {
        "coordinate_reference_system": metadata.coordinate_reference_system,
        "vertical_reference_system": metadata.vertical_reference_system,
        "data_classification": metadata.data_classification,
        "source_url": None,
        "plant_code": metadata.plant_code,
        "media_orientation_reference_system": metadata.media_orientation_reference_system,  # noqa: E501
        "additional_meta": {
            "mission_id": metadata.mission_id,
            "plant_name": metadata.plant_name,
            "mission_date": metadata.mission_date,
            "robot_id": metadata.robot_id,
        },
        "data": [
            {
                "folder": f"/{metadata.mission_id}",
                "files": [
                    {
                        "file_name": filename,
                        "timestamp": inspection.metadata.start_time,
                        "x": inspection.metadata.time_indexed_pose.pose.position.x,
                        "y": inspection.metadata.time_indexed_pose.pose.position.y,
                        "z": inspection.metadata.time_indexed_pose.pose.position.z,
                        "tag": inspection.metadata.tag_id,
                        "additional_media_metadata": {
                            "orientation": inspection.metadata.time_indexed_pose.pose.orientation.to_quat_array()  # noqa: E501
                        },
                    }
                ],
            }
        ],
    }

    return json.dumps(data, cls=EnhancedJSONEncoder, indent=4).encode()


def get_filename(
    mission_id: Any,
    inspection_type: str,
    inspection_id: UUID,
) -> str:
    return f"{mission_id}_{inspection_type}_{inspection_id}"


def get_inspection_type(inspection: Inspection) -> str:
    if isinstance(inspection, Image):
        return "image"
    elif isinstance(inspection, ThermalImage):
        return "thermal"
    elif isinstance(inspection, Video):
        return "video"
    elif isinstance(inspection, ThermalVideo):
        return "thermal_video"
    else:
        logging.getLogger("uploader").warning(
            f"Failed to upload inspection with ID: {inspection.id}\n "
            f"This was due to an inspection type ({type(inspection)}) "
            f"which is not supported by the uploader"
        )
        raise StorageException(
            f"Inspection type not supported by the uploader. Got {type(inspection)}"
        )



import importlib.resources as pkg_resources
from typing import List

from pydantic import BaseSettings, Field, validator


class Settings(BaseSettings):
    # Determines which robot package ISAR will attempt to import
    # Name must match with an installed python package in the local environment
    ROBOT_PACKAGE: str = Field(default="isar_robot")

    # Determines the local path in which results from missions are stored
    LOCAL_STORAGE_PATH: str = Field(default="./results")

    # Timeout in seconds for direct HTTP requests made through the RequestHandler
    REQUEST_TIMEOUT: int = Field(default=30)

    # Timeout in seconds for checking whether there is a message on a queue
    QUEUE_TIMEOUT: int = Field(default=10)

    # Sleep time for while loops in the finite state machine in seconds
    # The sleep is used to throttle the system on every iteration in the loop
    FSM_SLEEP_TIME: float = Field(default=0.1)

    # Location of JSON files containing predefined missions for the Local Planner to use
    PREDEFINED_MISSIONS_FOLDER: str = Field(
        default="src/isar/config/predefined_missions/"
    )

    # Name of default map transformation
    DEFAULT_MAP: str = Field(default="default_map")

    # Location of JSON files containing predefined maps
    MAPS_FOLDER: str = Field(default="src/isar/config/maps/")

    # Determines the number of state transitions that are kept in the log
    STATE_TRANSITIONS_LOG_LENGTH: int = Field(default=20)

    # Number of attempts to initiate a step before cancelling
    INITIATE_STEP_FAILURE_COUNTER_LIMIT: int = Field(default=10)

    # Number of attempts to stop the robot before giving up
    STOP_ROBOT_ATTEMPTS_LIMIT: int = Field(default=10)

    # Number of attempts to stop the robot before giving up
    UPLOAD_FAILURE_ATTEMPTS_LIMIT: int = Field(default=10)

    # Number of attempts to stop the robot before giving up
    UPLOAD_FAILURE_MAX_WAIT: int = Field(default=60)

    # FastAPI host
    API_HOST_VIEWED_EXTERNALLY: str = Field(default="0.0.0.0")

    # FastAPI port
    API_PORT: int = Field(default=3000)

    # Determines which mission planner module is used by ISAR
    # Options: [local echo]
    # Selecting a different mission planner module than local may require certain access
    # rights
    MISSION_PLANNER: str = Field(default="local")

    # Determines which task selector module is used by ISAR
    # Options: [sequential]
    TASK_SELECTOR: str = Field(default="sequential")

    # Determines which storage modules are used by ISAR
    # Multiple storage modules can be chosen
    # Each module will be called when storing results from inspections
    # Selecting a different storage module than local may require certain access rights
    STORAGE_LOCAL_ENABLED: bool = Field(default=True)
    STORAGE_BLOB_ENABLED: bool = Field(default=False)
    STORAGE_SLIMM_ENABLED: bool = Field(default=False)

    # Determines whether the MQTT publishing module should be enabled or not
    # The publishing module will attempt to connect to the MQTT broker configured in
    # "service_connections"
    # Options: [false true]
    MQTT_ENABLED: bool = Field(default=False)

    # Determines whether certificate based encryption will be used for the MQTT
    # communication.
    MQTT_SSL_ENABLED: bool = Field(default=True)

    # Determines whether authentication is enabled for the API or not
    # Enabling this requires certain resources available for OAuth2 authentication
    # Currently supported authentication is Azure AD
    # (https://github.com/Intility/fastapi-azure-auth)
    AUTHENTICATION_ENABLED: bool = Field(default=False)

    # Tenant ID for the Azure tenant with your Azure Active Directory application
    AZURE_TENANT_ID: str = Field(default="3aa4a235-b6e2-48d5-9195-7fcf05b459b0")

    # Client ID for the API client
    APP_CLIENT_ID: str = Field(default="68cca82d-84e7-495c-96b4-4c32509f2a46")

    # Client ID for the OpenAPI client
    OPENAPI_CLIENT_ID: str = Field(default="5f412c20-8c36-4c69-898f-d2b5051f5fb6")

    # MQTT username
    # The username and password is set by the MQTT broker and must be known in advance
    # The password should be set as an environment variable "MQTT_PASSWORD"
    # If the password is not set in the environment an empty string will be used
    MQTT_USERNAME: str = Field(default="isar")

    # MQTT host
    MQTT_HOST: str = Field(default="localhost")

    # MQTT port
    MQTT_PORT: int = Field(default=1883)

    # Keyvault name
    KEYVAULT: str = Field(default="EqRobotKeyVault")

    # URL to storage account for Azure Blob Storage
    BLOB_STORAGE_ACCOUNT_URL: str = Field(
        default="https://eqrobotdevstorage.blob.core.windows.net"
    )

    # Name of blob container in Azure Blob Storage [slimm test]
    BLOB_CONTAINER: str = Field(default="test")

    # Client ID for STID App Registration
    STID_CLIENT_ID: str = Field(default="1734406c-3449-4192-a50d-7c3a63d3f57d")

    # Scope for access to STID API
    STID_APP_SCOPE: str = Field(default=".default")

    # URL for STID endpoint
    STID_API_URL: str = Field(default="https://stidapi.equinor.com")

    # Plant name for the facility which STID should look for tags in
    STID_PLANT_NAME: str = Field(default="kaa")

    # Client ID for Echo App Registration
    ECHO_CLIENT_ID: str = Field(default="bf0b2569-e09c-42f0-8095-5a52a873eb7b")

    # Scope for access to Echo API
    ECHO_APP_SCOPE: str = Field(default=".default")

    # URL for Echo endpoint
    ECHO_API_URL: str = Field(default="https://echohubapi.equinor.com/api")

    # Client ID for SLIMM App Registration
    SLIMM_CLIENT_ID: str = Field(default="94c048cc-58e9-4570-85c0-4028c50ab6f3")

    # Scope for access to SLIMM Ingestion API
    SLIMM_APP_SCOPE: str = Field(default=".default")

    # URL for SLIMM endpoint
    SLIMM_API_URL: str = Field(
        default="https://slimmingestapitest.azurewebsites.net/SpatialIngest"
    )

    # Whether the results should be copied directly into the SLIMM datalake or only the
    # metadata
    COPY_FILES_TO_SLIMM_DATALAKE: bool = Field(default=False)

    # The configuration of this section is tightly coupled with the metadata that is
    # submitted with the results once they have been uploaded.

    # Four digit code indicating facility
    PLANT_CODE: str = Field(default="1320")

    # Name of the facility the robot is operating in
    PLANT_NAME: str = Field(default="Kårstø")

    # Shortname of the facility the robot is operating in
    PLANT_SHORT_NAME: str = Field(default="KAA")

    # Country the robot is operating in
    COUNTRY: str = Field(default="Norway")

    # Type of robot ISAR is monitoring
    ROBOT_TYPE: str = Field(default="robot")

    # Name or unique ID of robot
    ROBOT_ID: str = Field(default="R2-D2")

    # Data scheme the robot should adhere to
    # Options [DS0001]
    DATA_SCHEME: str = Field(default="DS0001")

    # Coordinate reference system of facility
    COORDINATE_REFERENCE_SYSTEM: str = Field(default="EQUINOR:4100001")

    # Vertical reference system of facility
    VERTICAL_REFERENCE_SYSTEM: str = Field(default="MSL")

    # Rotational representations of reported results
    # Options [quaternion]
    MEDIA_ORIENTATION_REFERENCE_SYSTEM: str = Field(default="quaternion")

    # Contractor who is responsible for robot missions
    CONTRACTOR: str = Field(default="equinor")

    # Timezone
    TIMEZONE: str = Field(default="UTC+00:00")

    # Data classification
    DATA_CLASSIFICATION: str = Field(default="internal")

    # List of MQTT Topics

    TOPIC_ISAR_ROBOT: str = Field(default="robot")

    TOPIC_ISAR_STATE: str = Field(default="state")

    TOPIC_ISAR_MISSION: str = Field(default="mission")

    TOPIC_ISAR_TASK: str = Field(default="task")

    TOPIC_ISAR_STEP: str = Field(default="step")

    API_LOG_LEVEL: str = Field(default="INFO")
    MAIN_LOG_LEVEL: str = Field(default="INFO")
    MQTT_LOG_LEVEL: str = Field(default="INFO")
    STATE_MACHINE_LOG_LEVEL: str = Field(default="INFO")
    UPLOADER_LOG_LEVEL: str = Field(default="INFO")
    CONSOLE_LOG_LEVEL: str = Field(default="INFO")
    URLLIB3_LOG_LEVEL: str = Field(default="WARNING")
    UVICORN_LOG_LEVEL: str = Field(default="WARNING")
    AZURE_LOG_LEVEL: str = Field(default="WARNING")

    LOG_LEVELS: dict = Field(default={})

    REQUIRED_ROLE: str = Field(default="Mission.Control")

    @validator("LOG_LEVELS", pre=True, always=True)
    def set_log_levels(cls, v, values) -> dict:
        return {
            "api": values["API_LOG_LEVEL"],
            "main": values["MAIN_LOG_LEVEL"],
            "mqtt": values["MQTT_LOG_LEVEL"],
            "state_machine": values["STATE_MACHINE_LOG_LEVEL"],
            "uploader": values["UPLOADER_LOG_LEVEL"],
            "console": values["CONSOLE_LOG_LEVEL"],
            "urllib3": values["URLLIB3_LOG_LEVEL"],
            "uvicorn": values["UVICORN_LOG_LEVEL"],
            "azure": values["AZURE_LOG_LEVEL"],
        }

    @validator(
        "TOPIC_ISAR_ROBOT",
        "TOPIC_ISAR_STATE",
        "TOPIC_ISAR_MISSION",
        "TOPIC_ISAR_TASK",
        "TOPIC_ISAR_STEP",
        pre=True,
        always=True,
    )
    def prefix_isar_topics(cls, v, values):
        return f"isar/{values['ROBOT_ID']}/{v}"

    class Config:
        with pkg_resources.path("isar.config", "settings.env") as path:
            package_path = path

        env_prefix = "ISAR_"
        env_file = package_path
        env_file_encoding = "utf-8"
        case_sensitive = True


settings = Settings()


class RobotSettings(BaseSettings):
    def __init__(self) -> None:
        try:
            with pkg_resources.path(
                f"{settings.ROBOT_PACKAGE}.config", "settings.env"
            ) as path:
                env_file_path = path
        except ModuleNotFoundError:
            env_file_path = None
        super().__init__(_env_file=env_file_path)

    CAPABILITIES: List[str] = Field(default=["drive_to_pose", "take_image"])

    class Config:
        env_file_encoding = "utf-8"
        case_sensitive = True


robot_settings = RobotSettings()



from typing import List, Tuple

from isar.models.mission import Mission


def is_robot_capable_of_mission(
    mission: Mission, robot_capabilities: List[str]
) -> Tuple[bool, List[str]]:
    success: bool = True
    missing_capabilities: List[str] = []
    for task in mission.tasks:
        for step in task.steps:
            if not step.type in robot_capabilities:
                success = False
                missing_capabilities.append(step.type)

    return success, missing_capabilities



import time
from collections import deque
from pathlib import Path
from threading import Thread
from typing import List

import pytest
import logging

from injector import Injector

from isar.mission_planner.local_planner import LocalPlanner
from isar.models.communication.queues.queues import Queues
from isar.models.mission import Mission, Task
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.services.utilities.threaded_request import ThreadedRequest
from isar.state_machine.state_machine import StateMachine, main
from isar.state_machine.states_enum import States
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.models.mission import DriveToPose, Step, TakeImage
from robot_interface.models.mission.status import StepStatus
from tests.mocks.pose import MockPose
from tests.mocks.robot_interface import MockRobot
from tests.mocks.step import MockStep
from pytest_mock import MockerFixture


class StateMachineThread(object):
    def __init__(self, injector) -> None:
        self.injector: Injector = injector
        self.state_machine: StateMachine = injector.get(StateMachine)
        self._thread: Thread = Thread(target=main, args=[injector])
        self._thread.daemon = True
        self._thread.start()


class UploaderThread(object):
    def __init__(self, injector) -> None:
        self.injector: Injector = injector
        self.uploader: Uploader = Uploader(
            upload_queue=self.injector.get(Queues).upload_queue,
            storage_handlers=injector.get(List[StorageInterface]),
        )
        self._thread: Thread = Thread(target=self.uploader.run)
        self._thread.daemon = True
        self._thread.start()


@pytest.fixture
def state_machine_thread(injector) -> StateMachineThread:
    return StateMachineThread(injector)


@pytest.fixture
def uploader_thread(injector) -> UploaderThread:
    return UploaderThread(injector=injector)


def get_mission():
    mission_reader: LocalPlanner = LocalPlanner()
    mission: Mission = mission_reader.read_mission_from_file(
        Path("./tests/test_data/test_mission_working.json")
    )
    return mission


def test_initial_off(state_machine):
    assert state_machine.state == "off"


def test_send_status(state_machine):
    state_machine.send_state_status()
    message = state_machine.queues.state.check()
    assert message == state_machine.current_state


def test_reset_state_machine(state_machine):
    state_machine.reset_state_machine()

    assert state_machine.current_step is None
    assert state_machine.current_task is None
    assert state_machine.current_mission is None


empty_mission: Mission = Mission([], None)


def test_state_machine_transitions(injector, state_machine_thread):
    step: Step = DriveToPose(pose=MockPose.default_pose)
    mission: Mission = Mission(tasks=[Task(steps=[step])])

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    scheduling_utilities.start_mission(mission=mission, initial_pose=None)

    time.sleep(3)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Initialize,
            States.InitiateStep,
            States.Monitor,
            States.InitiateStep,
            States.Idle,
        ]
    )
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )


def test_state_machine_failed_dependency(injector, state_machine_thread, mocker):
    drive_to_step: Step = DriveToPose(pose=MockPose.default_pose)
    inspection_step: Step = MockStep.take_image_in_coordinate_direction
    mission: Mission = Mission(tasks=[Task(steps=[drive_to_step, inspection_step])])

    mocker.patch.object(MockRobot, "step_status", return_value=StepStatus.Failed)

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    scheduling_utilities.start_mission(mission=mission, initial_pose=None)

    time.sleep(3)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Initialize,
            States.InitiateStep,
            States.Monitor,
            States.InitiateStep,
            States.Idle,
        ]
    )
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )


def test_state_machine_with_successful_collection(
    injector, state_machine_thread, uploader_thread
):
    storage_mock: StorageInterface = injector.get(List[StorageInterface])[0]

    step: TakeImage = MockStep.take_image_in_coordinate_direction
    mission: Mission = Mission(tasks=[Task(steps=[step])])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)

    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(3)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Initialize,
            States.InitiateStep,
            States.Monitor,
            States.InitiateStep,
            States.Idle,
        ]
    )
    expected_stored_items = 1
    assert len(storage_mock.stored_inspections) == expected_stored_items
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )


def test_state_machine_with_unsuccessful_collection(
    injector, mocker, state_machine_thread
):
    storage_mock: StorageInterface = injector.get(List[StorageInterface])[0]

    mocker.patch.object(MockRobot, "get_inspections", return_value=[])

    step: TakeImage = MockStep.take_image_in_coordinate_direction
    mission: Mission = Mission(tasks=[Task(steps=[step])])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)

    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(3)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Initialize,
            States.InitiateStep,
            States.Monitor,
            States.InitiateStep,
            States.Idle,
        ]
    )
    expected_stored_items = 0
    assert len(storage_mock.stored_inspections) == expected_stored_items
    print(state_machine_thread.state_machine.transitions_list)
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )


def test_state_machine_with_successful_mission_stop(
    injector: Injector,
    state_machine_thread: StateMachineThread,
    caplog: pytest.LogCaptureFixture,
):
    step: TakeImage = MockStep.take_image_in_coordinate_direction
    mission: Mission = Mission(tasks=[Task(steps=[step])])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    scheduling_utilities.stop_mission()
    expected = deque(
        [
            States.Idle,
            States.Initialize,
            States.InitiateStep,
            States.StopStep,
            States.Idle,
        ]
    )
    actual = state_machine_thread.state_machine.transitions_list
    unexpected_log = "Could not communicate request: Reached limit for stop attemps. Cancelled mission and transitioned to idle."
    assert unexpected_log not in caplog.text
    assert expected == actual


def test_state_machine_with_unsuccsessful_mission_stop(
    injector: Injector,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThread,
    caplog: pytest.LogCaptureFixture,
):
    step: TakeImage = MockStep.take_image_in_coordinate_direction
    mission: Mission = Mission(tasks=[Task(steps=[step])])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    mocker.patch.object(ThreadedRequest, "_is_thread_alive", return_value=True)
    scheduling_utilities.stop_mission()
    expected = deque(
        [
            States.Idle,
            States.Initialize,
            States.InitiateStep,
            States.StopStep,
            States.Idle,
        ]
    )
    actual = state_machine_thread.state_machine.transitions_list
    expected_log = "Could not communicate request: Reached limit for stop attemps. Cancelled mission and transitioned to idle."
    assert expected_log in caplog.text
    assert expected == actual



from abc import ABCMeta, abstractmethod
from queue import Queue
from threading import Thread
from typing import List, Sequence

from robot_interface.models.initialize import InitializeParams
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission import InspectionStep, Step, StepStatus


class RobotInterface(metaclass=ABCMeta):
    """Interface to communicate with robots."""

    @abstractmethod
    def initiate_step(self, step: Step) -> None:
        """Send a step to the robot and initiate the execution of the step

        Parameters
        ----------
        step : Step
            The step that should be initiated on the robot.

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the step is not initiated.

        """
        raise NotImplementedError

    @abstractmethod
    def step_status(self) -> StepStatus:
        """Gets the status of the currently active step on robot.

        Parameters
        ----------
        None

        Returns
        -------
        StepStatus
            Status of the execution of current step.

        Raises:
        ------
        RobotException
            If the step status can't be retrieved.

        """
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Stops the execution of the current step and stops the movement of the robot.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the robot is not stopped.

        """
        raise NotImplementedError

    @abstractmethod
    def get_inspections(self, step: InspectionStep) -> Sequence[Inspection]:
        """Return the inspections connected to the given step.

        Parameters
        ----------
        step : Step

        Returns
        -------
        Sequence[InspectionResult]
            List containing all the inspection results connected to the given step.

        Raises
        ------
        RobotException
            If the inspection results can't be retrieved.

        """
        raise NotImplementedError

    @abstractmethod
    def initialize(self, params: InitializeParams) -> None:
        """Initializes the robot. The initialization needed is robot dependent and the
        function can be a simple return statement if no initialization is needed for the
        robot.

        Parameters
        ----------
        params: InitializeParams

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the initialization failed

        """
        raise NotImplementedError

    @abstractmethod
    def get_telemetry_publishers(self, queue: Queue, robot_id: str) -> List[Thread]:
        """
        Set up telemetry publisher threads to publish regular updates for pose, battery
        level etc. from the robot to the MQTT broker. The publishers on the robot side
        will use the queue to pass messages to the MQTT Client on the ISAR side.

        The robot_id is passed to the robot to ensure the messages are published to the
        correct topics.

        Note that this functionality will only be utilized if MQTT is enabled in the
        settings.

        Returns
        -------
        List[Thread]
            List containing all threads that will be started to publish telemetry.

        """
        raise NotImplementedError



import pytest

from isar.mission_planner.mission_validator import is_robot_capable_of_mission
from tests.mocks.mission_definition import MockMissionDefinition


@pytest.mark.parametrize(
    "mission,capabilities,expected_return",
    [
        (MockMissionDefinition.default_mission, ["drive_to_pose"], False),
        (MockMissionDefinition.default_mission, ["drive_to_pose", "take_image"], True),
    ],
)
def test_is_robot_capable_of_mission(mission, capabilities, expected_return) -> None:
    return_value, missing_capabilities = is_robot_capable_of_mission(
        mission=mission, robot_capabilities=capabilities
    )
    assert return_value == expected_return



import logging
from typing import Dict, List, Optional, Union

from alitra import Frame, Pose, Position
from azure.identity import DefaultAzureCredential
from injector import inject
from requests import Response
from requests.exceptions import RequestException

from isar.config.predefined_poses.predefined_poses import predefined_poses
from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import (
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.models.mission import Mission, Task
from isar.services.auth.azure_credentials import AzureCredentials
from isar.services.service_connections.request_handler import RequestHandler
from isar.services.service_connections.stid.stid_service import StidService
from robot_interface.models.mission import DriveToPose, TakeImage, TakeThermalImage
from robot_interface.models.mission.step import TakeThermalVideo, TakeVideo


class EchoPlanner(MissionPlannerInterface):
    @inject
    def __init__(
        self,
        request_handler: RequestHandler,
        stid_service: StidService,
    ):
        self.request_handler: RequestHandler = request_handler
        self.stid_service: StidService = stid_service
        self.credentials: DefaultAzureCredential = (
            AzureCredentials.get_azure_credentials()
        )
        self.logger = logging.getLogger("api")

    def get_mission(self, mission_id: int) -> Mission:
        """
        Retrieve robot mission from echo mission planner with specified id.
        :param mission_id: Unique id of echo mission plan
        :return: Mission object
        """
        mission_plan: dict = self._mission_plan(mission_id)

        try:
            plan_items: List[dict] = mission_plan["planItems"]
        except KeyError as e:
            msg: str = "Echo request body don't contain expected keys"
            self.logger.error(msg)
            raise MissionPlannerError(msg) from e

        tasks: List[Task] = []

        for plan_item in plan_items:
            try:
                tag_id: str = plan_item["tag"]
                tag_position: Position = self._get_tag_position(tag_id=tag_id)
                drive_step: DriveToPose = self._create_drive_step(tag_id=tag_id)
                inspection_steps: List[
                    Union[TakeImage, TakeThermalImage, TakeVideo, TakeThermalVideo]
                ] = [
                    self._echo_sensor_to_isar_inspection_step(
                        sensor=sensor, tag_id=tag_id, tag_position=tag_position
                    )
                    for sensor in plan_item["sensorTypes"]
                ]
            except (ValueError, KeyError, RequestException) as e:
                self.logger.error(
                    f"Failed to create task with exception message: '{str(e)}'"
                )
                continue
            task: Task = Task(steps=[drive_step, *inspection_steps], tag_id=tag_id)
            tasks.append(task)

        if not tasks:
            raise MissionPlannerError("Empty mission")

        mission: Mission = Mission(tasks=tasks)

        return mission

    def _mission_plan(self, mission_id: int) -> dict:
        """
        Get mission plan from echo planner.
        :param mission_id: Unique id of echo mission plan
        """
        client_id: str = settings.ECHO_CLIENT_ID
        scope: str = settings.ECHO_APP_SCOPE
        request_scope: str = f"{client_id}/{scope}"

        token: str = self.credentials.get_token(request_scope).token

        url: str = f"{settings.ECHO_API_URL}/robots/robot-plan/{mission_id}"
        response: Response = self.request_handler.get(
            url=url,
            headers={"Authorization": f"Bearer {token}"},
        )

        return response.json()

    def _get_robot_pose(self, tag_id: str) -> Pose:
        """
        Retrieve robot pose corresponding to inspection of a given tag. For now, this is
        a temporary hard-coded solution.
        """
        try:
            predefined_pose: Pose = predefined_poses[tag_id]
        except KeyError:
            raise KeyError(f"Tag not in list of predefined poses: {tag_id}")
        if predefined_pose.frame == Frame("robot"):
            raise ValueError("Frame of predefined pose should be asset not robot")

        return predefined_pose

    def _get_tag_position(self, tag_id: str) -> Position:
        tag_position: Position = self.stid_service.tag_position(tag_id)

        return tag_position

    def _create_drive_step(self, tag_id: str) -> DriveToPose:
        robot_pose: Pose = self._get_robot_pose(tag_id=tag_id)

        drive_step: DriveToPose = DriveToPose(pose=robot_pose)
        return drive_step

    @staticmethod
    def _echo_sensor_to_isar_inspection_step(
        sensor: dict, tag_id: str, tag_position: Position
    ) -> Union[TakeImage, TakeThermalImage, TakeVideo, TakeThermalVideo]:
        sensor_type: str = sensor["sensorTypeKey"]
        duration: Optional[float] = sensor["timeInSeconds"]
        inspection: Union[TakeImage, TakeThermalImage, TakeVideo, TakeThermalVideo]
        if sensor_type == "Picture":
            inspection = TakeImage(target=tag_position)
        elif sensor_type == "Video":
            inspection = TakeVideo(target=tag_position, duration=duration)
        elif sensor_type == "ThermicPicture":
            inspection = TakeThermalImage(target=tag_position)
        elif sensor_type == "ThermicVideo":
            inspection = TakeThermalVideo(target=tag_position, duration=duration)
        else:
            raise ValueError(f"No step supported for sensor_type {sensor_type}")
        inspection.tag_id = tag_id
        return inspection



import json
import logging
import os
from datetime import datetime
from queue import Empty, Queue

import backoff
from paho.mqtt import client as mqtt
from paho.mqtt.client import Client

from isar.config.settings import settings
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class MqttClient(MqttClientInterface):
    def __init__(self, mqtt_queue: Queue) -> None:
        self.logger = logging.getLogger("mqtt_client")
        self.logger.setLevel("INFO")
        self.mqtt_queue: Queue = mqtt_queue

        username: str = settings.MQTT_USERNAME
        password: str = ""
        try:
            password = os.environ["ISAR_MQTT_PASSWORD"]
        except KeyError:
            self.logger.warning(
                "Failed to retrieve ISAR_MQTT_PASSWORD from environment. Attempting "
                "with empty string as password."
            )

        self.host: str = settings.MQTT_HOST

        # Fix for mqtt running on localhost in docker
        if "IS_DOCKER" in os.environ and self.host == "localhost":
            self.host = "host.docker.internal"

        self.port: int = settings.MQTT_PORT

        self.client: Client = Client()

        self.client.enable_logger(logger=self.logger)

        dirname = os.path.dirname(__file__)

        if settings.MQTT_SSL_ENABLED:
            cert_path = os.path.join(dirname, "../../../config/certs/ca-cert.pem")
            self.client.tls_set(ca_certs=cert_path)

        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect

        self.client.username_pw_set(username=username, password=password)

    def run(self) -> None:
        self.connect(host=self.host, port=self.port)
        self.client.loop_start()
        payload: str = json.dumps(
            {
                "robot_id": settings.ROBOT_ID,
                "host": settings.API_HOST_VIEWED_EXTERNALLY,
                "port": settings.API_PORT,
                "timestamp": datetime.utcnow(),
            },
            cls=EnhancedJSONEncoder,
        )

        self.publish(
            topic=settings.TOPIC_ISAR_ROBOT,
            payload=payload,
            retain=True,
        )

        while True:
            if not self.client.is_connected():
                continue
            try:
                topic, payload, qos, retain = self.mqtt_queue.get(timeout=1)
            except Empty:
                continue

            self.publish(topic=topic, payload=payload, qos=qos, retain=retain)

    def on_connect(self, client, userdata, flags, rc):
        self.logger.info("Connection returned result: " + mqtt.connack_string(rc))

    def on_disconnect(self, client, userdata, rc):
        if rc != 0:
            self.logger.warning("Unexpected disconnection from MQTT Broker")

    @staticmethod
    def _on_success(data: dict) -> None:
        logging.getLogger("mqtt_client").info("Connected to MQTT Broker")
        logging.getLogger("mqtt_client").debug(
            f"Elapsed time: {data['elapsed']}, Tries: {data['tries']}"
        )

    @staticmethod
    def _on_backoff(data: dict) -> None:
        logging.getLogger("mqtt_client").warning(
            f"Failed to connect, retrying in {data['wait']} seconds"
        )

    @staticmethod
    def _on_giveup(data: dict) -> None:
        logging.getLogger("mqtt_client").error(
            "Failed to connect to MQTT Broker within set backoff strategy."
        )

    @backoff.on_exception(
        backoff.expo,
        ConnectionRefusedError,
        max_time=300,
        on_success=_on_success,
        on_backoff=_on_backoff,
        on_giveup=_on_giveup,
    )
    def connect(self, host: str, port: int) -> None:
        self.logger.info("Attempting to connect to MQTT Broker")
        self.logger.debug(f"Host: {host}, Port: {port}")
        self.client.connect(host=host, port=port)

    def publish(self, topic: str, payload: str, qos: int = 0, retain: bool = False):
        self.logger.debug(f"Publishing message to topic: {topic}")
        self.client.publish(topic=topic, payload=payload, qos=qos, retain=retain)



import json
import logging
from dataclasses import is_dataclass
from logging import Logger
from pathlib import Path
from typing import Any, Optional

from dacite import Config, from_dict

logger: Logger = logging.getLogger("api")


class BaseReader:
    @staticmethod
    def read_json(location: Path) -> dict:
        with open(location) as json_file:
            return json.load(json_file)

    @staticmethod
    def dict_to_dataclass(
        dataclass_dict: dict,
        target_dataclass: Any,
        cast_config: list = [],
        strict_config: bool = False,
    ) -> Optional[Any]:
        if not is_dataclass(target_dataclass):
            raise BaseReaderError("{target_dataclass} is not a dataclass")
        generated_dataclass = from_dict(
            data_class=target_dataclass,
            data=dataclass_dict,
            config=Config(cast=cast_config, strict=strict_config),
        )
        return generated_dataclass


class BaseReaderError(Exception):
    pass



from queue import Empty

import pytest

from isar.models.communication.queues import Queues, StatusQueue


class TestQueues:
    def test_queues(self):
        queues = Queues()
        assert queues.start_mission is not None
        assert (
            queues.start_mission.input is not None
            and queues.start_mission.input.maxsize == 1
        )
        assert (
            queues.start_mission.output is not None
            and queues.start_mission.output.maxsize == 1
        )
        assert queues.stop_mission is not None
        assert (
            queues.stop_mission.input is not None
            and queues.stop_mission.input.maxsize == 1
        )
        assert (
            queues.stop_mission.output is not None
            and queues.stop_mission.output.maxsize == 1
        )


def test_staus_queue_empty():
    status_queue: StatusQueue = StatusQueue()
    with pytest.raises(Empty):
        status_queue.check()


def test_status_queue_check():
    status_queue: StatusQueue = StatusQueue()
    status_queue.update("Test")
    assert status_queue.check() == "Test"
    assert status_queue._qsize() == 1


def test_status_queue_update():
    status_queue: StatusQueue = StatusQueue()
    status_queue.update("Test")
    status_queue.update("New Test")
    assert status_queue._qsize() == 1
    assert status_queue.check() == "New Test"



from typing import List

from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.storage.storage_interface import StorageException, StorageInterface
from robot_interface.models.inspection.inspection import Inspection


class StorageMock(StorageInterface):
    will_fail: bool = False

    def __init__(self) -> None:
        self.stored_inspections: List[Inspection] = []

    def store(self, inspection: Inspection, metadata: MissionMetadata):
        if self.will_fail:
            raise StorageException("Mock failed on purpose")
        self.stored_inspections.append(inspection)

    def blob_exists(self, inspection: Inspection) -> bool:
        return inspection in self.stored_inspections



from dataclasses import dataclass, field
from datetime import datetime
from typing import Iterator, List, Optional, Union
from uuid import UUID, uuid4

from isar.config.settings import settings
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from robot_interface.models.mission import (
    InspectionStep,
    STEPS,
    MotionStep,
    Step,
    StepStatus,
)
from robot_interface.models.mission.step import DriveToPose

from .status import MissionStatus, TaskStatus


@dataclass
class Task:
    steps: List[STEPS]
    status: TaskStatus = field(default=TaskStatus.NotStarted, init=False)
    tag_id: Optional[str] = field(default=None)
    id: UUID = field(default_factory=uuid4, init=False)
    _iterator: Iterator = None

    def next_step(self) -> Step:
        step: Step = next(self._iterator)
        while step.status != StepStatus.NotStarted:
            step = next(self._iterator)
        return step

    def is_finished(self) -> bool:
        for step in self.steps:
            if step.status is StepStatus.Failed and isinstance(step, MotionStep):
                # One motion step has failed meaning the task as a whole should be
                # aborted
                self.status = TaskStatus.Failed
                return True

            elif (step.status is StepStatus.Failed) and isinstance(
                step, InspectionStep
            ):
                # It should be possible to perform several inspections per task. If
                # one out of many inspections fail the task is considered as
                # partially successful.
                self.status = TaskStatus.PartiallySuccessful
                continue

            elif step.status is StepStatus.Successful:
                # The task is complete once all steps are completed
                continue
            else:
                # Not all steps have been completed yet
                return False

        # Check if the task has been marked as partially successful by having one or
        # more inspection steps fail
        if self.status is not TaskStatus.PartiallySuccessful:
            # All steps have been completed
            self.status = TaskStatus.Successful

        # Set the task to failed if all inspection steps failed
        elif self._all_inspection_steps_failed():
            self.status = TaskStatus.Failed

        return True

    def reset_task(self):
        for step in self.steps:
            if isinstance(step, DriveToPose):
                step.status = StepStatus.NotStarted
            elif (
                isinstance(step, InspectionStep)
                and step.status == StepStatus.InProgress
            ):
                step.status = StepStatus.NotStarted
        self._iterator = iter(self.steps)

    def _all_inspection_steps_failed(self) -> bool:
        for step in self.steps:
            if isinstance(step, MotionStep):
                continue
            elif step.status is not StepStatus.Failed:
                return False

        return True

    def __post_init__(self) -> None:
        if self._iterator is None:
            self._iterator = iter(self.steps)

    def api_response_dict(self):
        return {
            "id": self.id,
            "tag_id": self.tag_id,
            "steps": list(
                map(lambda x: {"id": x.id, "type": x.__class__.__name__}, self.steps)
            ),
        }


@dataclass
class Mission:
    tasks: List[Task]
    id: Union[UUID, int, str, None] = None
    status: MissionStatus = MissionStatus.NotStarted
    metadata: MissionMetadata = None

    def set_unique_id_and_metadata(self) -> None:
        self._set_unique_id()
        self.metadata = MissionMetadata(mission_id=self.id)

    def _set_unique_id(self) -> None:
        plant_short_name: str = settings.PLANT_SHORT_NAME
        robot_id: str = settings.ROBOT_ID
        now: datetime = datetime.utcnow()
        self.id = (
            f"{plant_short_name.upper()}{robot_id.upper()}"
            f"{now.strftime('%d%m%Y%H%M%S%f')[:-3]}"
        )

    def __post_init__(self) -> None:
        if self.id is None:
            self._set_unique_id()

        if self.metadata is None:
            self.metadata = MissionMetadata(mission_id=self.id)

    def api_response_dict(self) -> dict:
        return {
            "id": self.id,
            "tasks": [task.api_response_dict() for task in self.tasks],
        }



import logging
import traceback

from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceNotFoundError,
)
from azure.identity import ClientSecretCredential, DefaultAzureCredential
from azure.keyvault.secrets import KeyVaultSecret, SecretClient

from isar.config.keyvault.keyvault_error import KeyvaultError


class Keyvault:
    def __init__(
        self,
        keyvault_name: str,
        client_id: str = None,
        client_secret: str = None,
        tenant_id: str = None,
    ):
        self.url = "https://" + keyvault_name + ".vault.azure.net"
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.logger = logging.getLogger("API")

    def get_secret(self, secret_name: str) -> KeyVaultSecret:
        secret_client: SecretClient = self.get_secret_client()
        try:
            secret: KeyVaultSecret = secret_client.get_secret(name=secret_name)
        except ResourceNotFoundError:
            self.logger.error("Secret was not found in keyvault.")
            traceback.print_exc()
            raise KeyvaultError  # type: ignore
        except HttpResponseError:
            self.logger.error(
                "An error occurred while retrieving a secret from keyvault."
            )
            traceback.print_exc()
            raise KeyvaultError  # type: ignore

        return secret

    def set_secret(self, secret_name: str, secret_value) -> None:
        secret_client: SecretClient = self.get_secret_client()
        try:
            secret_client.set_secret(name=secret_name, value=secret_value)
        except HttpResponseError:
            self.logger.error("An error occurred while setting a secret in keyvault.")
            traceback.print_exc()
            raise KeyvaultError  # type: ignore

    def get_secret_client(self):
        try:
            if self.client_id and self.client_secret and self.tenant_id:
                credential: ClientSecretCredential = ClientSecretCredential(
                    tenant_id=self.tenant_id,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                )
            else:
                credential: DefaultAzureCredential = DefaultAzureCredential()
        except ClientAuthenticationError:
            self.logger.error("Failed to authenticate to Azure.")
            traceback.print_exc()
            raise KeyvaultError

        secret_client = SecretClient(vault_url=self.url, credential=credential)
        return secret_client



from pathlib import Path

import pytest
from alitra import Frame, Orientation, Pose, Position

from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.models.mission import Mission, Task
from robot_interface.models.mission import TakeThermalImage
from robot_interface.models.mission.step import DriveToPose, Step, TakeImage


@pytest.mark.parametrize(
    "mission_path",
    [
        Path("./tests/test_data/test_mission_working_no_tasks.json"),
        Path("./tests/test_data/test_mission_working.json"),
    ],
)
def test_get_mission(mission_reader, mission_path):
    output: Mission = mission_reader.read_mission_from_file(mission_path)
    assert isinstance(output, Mission)


def test_read_mission_from_file(mission_reader):
    expected_step_1 = DriveToPose(
        pose=Pose(
            position=Position(-2, -2, 0, Frame("asset")),
            orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
            frame=Frame("asset"),
        )
    )
    task_1: Task = Task(steps=[expected_step_1])

    expected_step_2 = DriveToPose(
        pose=Pose(
            position=Position(-2, 2, 0, Frame("asset")),
            orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
            frame=Frame("asset"),
        )
    )
    expected_step_3 = TakeImage(target=Position(2, 2, 0, Frame("robot")))
    task_2: Task = Task(steps=[expected_step_2, expected_step_3])

    expected_step_4 = DriveToPose(
        pose=Pose(
            position=Position(2, 2, 0, Frame("asset")),
            orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
            frame=Frame("asset"),
        )
    )
    expected_step_5 = TakeImage(target=Position(2, 2, 0, Frame("robot")))
    task_3: Task = Task(steps=[expected_step_4, expected_step_5])

    expected_step_6 = DriveToPose(
        pose=Pose(
            position=Position(0, 0, 0, Frame("asset")),
            orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
            frame=Frame("asset"),
        )
    )
    task_4: Task = Task(steps=[expected_step_6])

    expected_tasks = [task_1, task_2, task_3, task_4]
    expected_mission: Mission = Mission(tasks=expected_tasks)
    mission: Mission = mission_reader.read_mission_from_file(
        Path("./tests/test_data/test_mission_working.json")
    )
    assert (
        expected_mission.metadata.coordinate_reference_system
        == mission.metadata.coordinate_reference_system
    )
    assert (
        expected_mission.metadata.vertical_reference_system
        == mission.metadata.vertical_reference_system
    )
    assert (
        expected_mission.metadata.data_classification
        == mission.metadata.data_classification
    )
    for (expected_task, task) in zip(expected_tasks, mission.tasks):
        for (expected_step, step) in zip(expected_task.steps, task.steps):
            if isinstance(expected_step, DriveToPose) and isinstance(step, DriveToPose):
                assert expected_step.pose == step.pose
            if isinstance(expected_step, TakeImage) and isinstance(step, TakeImage):
                assert expected_step.target == step.target


@pytest.mark.parametrize(
    "mission_path",
    [
        (Path("./tests/test_data/no_file.json")),
        (Path("./tests/test_data/test_mission_not_working.json")),
    ],
)
def test_get_invalid_mission(mission_reader, mission_path):
    with pytest.raises(Exception):
        mission_reader.read_mission_from_file(mission_path)


def test_get_mission_by_id(mission_reader):
    output = mission_reader.get_mission(1)
    assert isinstance(output, Mission)


def test_get_mission_by_invalid_id(mission_reader):
    with pytest.raises(MissionPlannerError):
        mission_reader.get_mission(12345)


def test_valid_predefined_missions_files(mission_reader):
    # Checks that the predefined mission folder contains only valid missions!
    mission_list_dict = mission_reader.get_predefined_missions()
    predefined_mission_folder = Path(settings.PREDEFINED_MISSIONS_FOLDER)
    assert len(list(predefined_mission_folder.glob("*.json"))) == len(
        list(mission_list_dict)
    )
    for file in predefined_mission_folder.glob("*.json"):
        path_to_file = predefined_mission_folder.joinpath(file.name)
        mission: Mission = mission_reader.read_mission_from_file(path_to_file)
        assert mission is not None


def test_thermal_image_step(mission_reader):
    mission_path: Path = Path("./tests/test_data/test_thermal_image_mission.json")
    output: Mission = mission_reader.read_mission_from_file(mission_path)

    step: Step = output.tasks[0].steps[0]

    assert isinstance(step, TakeThermalImage)
    assert hasattr(step, "target")
    assert step.type == "take_thermal_image"
    assert hasattr(step, "id")
    assert hasattr(step, "tag_id")



from queue import Queue


class QueueIO:
    """
    Creates input and output queue. The queues are defined such that the input is from
    api to state machine while the output is from state machine to api.
    """

    def __init__(self, input_size: int = 0, output_size: int = 0):
        self.input: Queue = Queue(maxsize=input_size)
        self.output: Queue = Queue(maxsize=output_size)



from .inspection import (
    ImageMetadata,
    TimeIndexedPose,
    InspectionMetadata,
    ImageMetadata,
    ThermalImageMetadata,
    VideoMetadata,
    ThermalVideoMetadata,
    Inspection,
    Image,
    ThermalImage,
    ThermalVideo,
    Video,
)



import time
from datetime import datetime
from threading import Thread
from typing import List, Tuple

import pytest
from alitra import Frame, Orientation, Pose, Position
from injector import Injector

from isar.models.communication.queues.queues import Queues
from isar.models.mission import Mission
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.models.inspection.inspection import (
    ImageMetadata,
    Inspection,
    TimeIndexedPose,
)

MISSION_ID = "some-mission-id"
ARBITRARY_IMAGE_METADATA = ImageMetadata(
    datetime.now(),
    TimeIndexedPose(
        Pose(
            Position(0, 0, 0, Frame("asset")),
            Orientation(x=0, y=0, z=0, w=1, frame=Frame("asset")),
            Frame("asset"),
        ),
        datetime.now(),
    ),
    file_type="jpg",
)
DATA_BYTES: bytes = b"Lets say this is some image data"


class UploaderThread(object):
    def __init__(self, injector) -> None:
        self.injector: Injector = injector
        self.uploader: Uploader = Uploader(
            upload_queue=self.injector.get(Queues).upload_queue,
            storage_handlers=injector.get(List[StorageInterface]),
        )
        self._thread: Thread = Thread(target=self.uploader.run)
        self._thread.daemon = True
        self._thread.start()


@pytest.fixture
def uploader_thread(injector) -> UploaderThread:
    return UploaderThread(injector=injector)


def test_should_upload_from_queue(uploader_thread):
    mission: Mission = Mission([])
    inspection: Inspection = Inspection(ARBITRARY_IMAGE_METADATA)
    message: Tuple[Inspection, MissionMetadata] = (
        inspection,
        mission.metadata,
    )
    uploader_thread.uploader.upload_queue.put(message)
    time.sleep(1)
    assert uploader_thread.uploader.storage_handlers[0].blob_exists(inspection)


def test_should_retry_failed_upload_from_queue(uploader_thread, mocker):
    mission: Mission = Mission([])
    inspection: Inspection = Inspection(ARBITRARY_IMAGE_METADATA)
    message: Tuple[Inspection, MissionMetadata] = (
        inspection,
        mission.metadata,
    )

    # Need it to fail so that it retries
    uploader_thread.uploader.storage_handlers[0].will_fail = True
    uploader_thread.uploader.upload_queue.put(message)
    time.sleep(1)

    # Should not upload, instead raise StorageException
    assert not uploader_thread.uploader.storage_handlers[0].blob_exists(inspection)
    uploader_thread.uploader.storage_handlers[0].will_fail = False
    time.sleep(3)

    # After 3 seconds, it should have retried and now it should be successful
    assert uploader_thread.uploader.storage_handlers[0].blob_exists(inspection)



import logging
import time
from copy import deepcopy
from typing import TYPE_CHECKING, Callable, Optional, Sequence, Tuple

from injector import inject
from transitions import State

from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions import RobotException
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission import InspectionStep, Step, StepStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Monitor(State):
    @inject
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="monitor", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine

        self.logger = logging.getLogger("state_machine")
        self.step_status_thread: Optional[ThreadedRequest] = None

    def start(self):
        self.state_machine.update_state()
        self._run()

    def stop(self):
        if self.step_status_thread:
            self.step_status_thread.wait_for_thread()
        self.step_status_thread = None

    def _run(self):
        transition: Callable
        while True:
            if self.state_machine.should_stop_mission():
                transition = self.state_machine.stop
                break

            if self.state_machine.should_pause_mission():
                transition = self.state_machine.pause
                break

            if not self.step_status_thread:
                self.step_status_thread = ThreadedRequest(
                    self.state_machine.robot.step_status
                )
                self.step_status_thread.start_thread()

            try:
                step_status: StepStatus = self.step_status_thread.get_output()
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue
            except RobotException:
                step_status = StepStatus.Failed

            self.state_machine.current_step.status = step_status

            if self._step_finished(step=self.state_machine.current_step):
                self._process_finished_step(step=self.state_machine.current_step)
                transition = self.state_machine.step_finished
                break

            self.step_status_thread = None
            time.sleep(self.state_machine.sleep_time)

        transition()

    def _queue_inspections_for_upload(self, current_step: InspectionStep):
        try:
            inspections: Sequence[
                Inspection
            ] = self.state_machine.robot.get_inspections(step=current_step)
        except Exception as e:
            self.logger.error(
                f"Error getting inspections for step {str(current_step.id)[:8]}: {e}"
            )
            return

        if not inspections:
            self.logger.warning(
                f"No inspection data retrieved for step {str(current_step.id)[:8]}"
            )

        # A deepcopy is made to freeze the metadata before passing it to another thread
        # through the queue
        mission_metadata: MissionMetadata = deepcopy(
            self.state_machine.current_mission.metadata
        )

        for inspection in inspections:
            inspection.metadata.tag_id = current_step.tag_id

            message: Tuple[Inspection, MissionMetadata] = (
                inspection,
                mission_metadata,
            )
            self.state_machine.queues.upload_queue.put(message)
            self.logger.info(f"Inspection: {str(inspection.id)[:8]} queued for upload")

    def _step_finished(self, step: Step) -> bool:
        finished: bool = False
        if step.status == StepStatus.Failed:
            self.logger.warning(f"Step: {str(step.id)[:8]} failed")
            finished = True
        elif step.status == StepStatus.Successful:
            self.logger.info(
                f"{type(step).__name__} step: {str(step.id)[:8]} completed"
            )
            finished = True
        return finished

    def _process_finished_step(self, step: Step) -> None:
        if step.status == StepStatus.Successful and isinstance(step, InspectionStep):
            self._queue_inspections_for_upload(current_step=step)



from isar.models.mission import Mission, Task
from tests.mocks.step import MockStep


class MockMissionDefinition:
    default_mission = Mission(
        id="default_mission",
        tasks=[
            Task(
                steps=[MockStep.take_image_in_coordinate_direction, MockStep.drive_to]
            ),
            Task(
                steps=[
                    MockStep.take_image_in_coordinate_direction,
                    MockStep.take_image_in_coordinate_direction,
                ]
            ),
        ],
    )



import pytest
from alitra import Frame, Position
from azure.identity._credentials.default import DefaultAzureCredential

from isar.services.service_connections.request_handler import RequestHandler
from tests.mocks.request import MockRequests
from tests.mocks.token import MockToken


@pytest.mark.parametrize(
    "tag, expected_position, mock_return",
    [
        (
            "334-LD-0225",
            Position(x=20196.2000, y=5248.4740, z=15.2080, frame=Frame("asset")),
            MockRequests(
                json_data={
                    "xCoordinate": 20196200.0,
                    "yCoordinate": 5248474.0,
                    "zCoordinate": 15208.0,
                }
            ),
        ),
    ],
)
def test_get_position(stid_service, mocker, tag, expected_position, mock_return):
    mocker.patch.object(DefaultAzureCredential, "get_token", return_value=MockToken())
    mocker.patch.object(RequestHandler, "get", return_value=mock_return)
    position: Position = stid_service.tag_position(tag=tag)
    assert position == expected_position



import pytest
from alitra import Frame, Position
from requests import RequestException

from isar.mission_planner.echo_planner import EchoPlanner
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.models.mission import Mission
from isar.services.service_connections.stid.stid_service import StidService
from robot_interface.models.mission import DriveToPose, TakeImage, TakeThermalImage
from robot_interface.models.mission.step import TakeThermalVideo, TakeVideo


@pytest.mark.parametrize(
    "id, mock_return, mock_stid_side_effect, mock_stid, expected_return",
    [
        (
            76,
            {
                "robotPlanId": 76,
                "planItems": [
                    {
                        "planItemId": 1227,
                        "tag": "313-PA-101A",
                        "sensorTypes": [
                            {"sensorTypeKey": "ThermicPicture", "timeInSeconds": None},
                            {"sensorTypeKey": "Picture", "timeInSeconds": None},
                        ],
                        "sortingOrder": 0,
                        "robotPlanId": 76,
                    },
                ],
            },
            None,
            Position(x=1, y=1, z=0, frame=Frame("asset")),
            {
                "steps": 3,
                "step_1_type": DriveToPose,
                "step_2_type": TakeThermalImage,
            },
        ),
        (
            76,
            {
                "robotPlanId": 76,
                "planItems": [
                    {
                        "planItemId": 1227,
                        "tag": "355-LD-1003",
                        "sensorTypes": [
                            {"sensorTypeKey": "Picture", "timeInSeconds": None}
                        ],
                        "sortingOrder": 0,
                        "robotPlanId": 76,
                    },
                ],
            },
            None,
            Position(x=1, y=1, z=0, frame=Frame("asset")),
            {
                "steps": 2,
                "step_1_type": DriveToPose,
                "step_2_type": TakeImage,
            },
        ),
        (
            76,
            {
                "robotPlanId": 76,
                "planItems": [
                    {
                        "planItemId": 1227,
                        "tag": "355-LD-1003",
                        "sensorTypes": [
                            {"sensorTypeKey": "Video", "timeInSeconds": 10}
                        ],
                        "sortingOrder": 0,
                        "robotPlanId": 76,
                    },
                ],
            },
            None,
            Position(x=1, y=1, z=0, frame=Frame("asset")),
            {
                "steps": 2,
                "step_1_type": DriveToPose,
                "step_2_type": TakeVideo,
            },
        ),
        (
            76,
            {
                "robotPlanId": 76,
                "planItems": [
                    {
                        "planItemId": 1227,
                        "tag": "355-LD-1003",
                        "sensorTypes": [
                            {"sensorTypeKey": "ThermicVideo", "timeInSeconds": 10}
                        ],
                        "sortingOrder": 0,
                        "robotPlanId": 76,
                    },
                ],
            },
            None,
            Position(x=1, y=1, z=0, frame=Frame("asset")),
            {
                "steps": 2,
                "step_1_type": DriveToPose,
                "step_2_type": TakeThermalVideo,
            },
        ),
    ],
)
def test_get_echo_mission(
    echo_service,
    mocker,
    id,
    mock_return,
    mock_stid_side_effect,
    mock_stid,
    expected_return,
):
    mocker.patch.object(EchoPlanner, "_mission_plan", return_value=mock_return)
    mocker.patch.object(
        StidService,
        "tag_position",
        return_value=mock_stid,
        side_effect=mock_stid_side_effect,
    )
    mission: Mission = echo_service.get_mission(mission_id=id)
    assert len(mission.tasks[0].steps) == expected_return["steps"]
    if not len(mission.tasks) == 0:
        assert isinstance(mission.tasks[0].steps[0], expected_return["step_1_type"])
        assert isinstance(mission.tasks[0].steps[1], expected_return["step_2_type"])


def test_get_echo_mission_raises_when_empty_mission(echo_service, mocker):
    mock_return = {
        "robotPlanId": 76,
        "planItems": [
            {
                "planItemId": 1227,
                "tag": "wrong_tag",
                "senorTypes": [{"sensorTypeKey": "Picture"}],
                "sortingOrder": 0,
                "robotPlanId": 76,
            },
        ],
    }
    mocker.patch.object(EchoPlanner, "_mission_plan", return_value=mock_return)
    mocker.patch.object(
        StidService,
        "tag_position",
        side_effect=RequestException,
    )
    with pytest.raises(MissionPlannerError):
        echo_service.get_mission(mission_id=76)



import logging
import time
from typing import TYPE_CHECKING, Callable, Optional

from injector import inject
from transitions import State

from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions import RobotException

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Initialize(State):
    @inject
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="initialize", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine

        self.logger = logging.getLogger("state_machine")
        self.initialize_thread: Optional[ThreadedRequest] = None

    def start(self):
        self.state_machine.update_state()
        self._run()

    def stop(self):
        if self.initialize_thread:
            self.initialize_thread.wait_for_thread()
        self.initialize_thread = None

    def _run(self):
        transition: Callable
        while True:
            if not self.initialize_thread:
                self.initialize_thread = ThreadedRequest(
                    self.state_machine.robot.initialize
                )
                self.initialize_thread.start_thread(
                    self.state_machine.get_initialize_params()
                )

            try:
                self.initialize_thread.get_output()
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue
            except RobotException as e:
                self.logger.error(f"Initialization of robot failed. Error: {e}")
                transition = self.state_machine.initialization_failed
                break

            transition = self.state_machine.initialization_successful
            break
        transition()



import time
from abc import ABCMeta, abstractmethod
from queue import Queue
from typing import Callable, Tuple

from robot_interface.models.exceptions.robot_exceptions import (
    RobotInvalidTelemetryException,
)


class MqttClientInterface(metaclass=ABCMeta):
    @abstractmethod
    def publish(
        self, topic: str, payload: str, qos: int = 0, retain: bool = False
    ) -> None:
        """
        Parameters
        ----------
        topic : string
            MQTT topic to publish to
        payload : string
            Payload to send to publish on the topic
        qos : integer
            Quality of Service
        retain : boolean
            Retain on topic

        Returns
        -------

        """
        pass


class MqttPublisher(MqttClientInterface):
    def __init__(self, mqtt_queue: Queue) -> None:
        self.mqtt_queue: Queue = mqtt_queue

    def publish(
        self, topic: str, payload: str, qos: int = 0, retain: bool = False
    ) -> None:
        queue_message: Tuple[str, str, int, bool] = (topic, payload, qos, retain)
        self.mqtt_queue.put(queue_message)


class MqttTelemetryPublisher(MqttClientInterface):
    def __init__(
        self,
        mqtt_queue: Queue,
        telemetry_method: Callable,
        topic: str,
        interval: float,
        qos: int = 0,
        retain: bool = False,
    ) -> None:
        self.mqtt_queue: Queue = mqtt_queue
        self.telemetry_method: Callable = telemetry_method
        self.topic: str = topic
        self.interval: float = interval
        self.qos: int = qos
        self.retain: bool = retain

    def run(self, robot_id: str) -> None:
        while True:
            try:
                payload: str = self.telemetry_method(robot_id)
            except RobotInvalidTelemetryException:
                continue

            self.publish(
                topic=self.topic, payload=payload, qos=self.qos, retain=self.retain
            )
            time.sleep(self.interval)

    def publish(
        self, topic: str, payload: str, qos: int = 0, retain: bool = False
    ) -> None:
        queue_message: Tuple[str, str, int, bool] = (topic, payload, qos, retain)
        self.mqtt_queue.put(queue_message)



import logging

from azure.identity import DefaultAzureCredential
from injector import inject
from requests import HTTPError, RequestException
from requests_toolbelt import MultipartEncoder

from isar.config.settings import settings
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.services.auth.azure_credentials import AzureCredentials
from isar.services.service_connections.request_handler import RequestHandler
from isar.storage.storage_interface import StorageException, StorageInterface
from isar.storage.utilities import get_filename, get_inspection_type
from robot_interface.models.inspection.inspection import Inspection


class SlimmStorage(StorageInterface):
    @inject
    def __init__(self, request_handler: RequestHandler) -> None:
        self.request_handler: RequestHandler = request_handler
        self.logger = logging.getLogger("uploader")

        self.credentials: DefaultAzureCredential = (
            AzureCredentials.get_azure_credentials()
        )

        client_id: str = settings.SLIMM_CLIENT_ID
        scope: str = settings.SLIMM_APP_SCOPE
        self.request_scope: str = f"{client_id}/{scope}"

        self.url: str = settings.SLIMM_API_URL

    def store(self, inspection: Inspection, metadata: MissionMetadata):
        token: str = self.credentials.get_token(self.request_scope).token

        request_url: str = f"{self.url}/UploadSingleFile"

        inspection_type: str = get_inspection_type(inspection=inspection)
        filename: str = get_filename(
            mission_id=metadata.mission_id,
            inspection_type=inspection_type,
            inspection_id=inspection.id,
        )
        filename_with_ending: str = f"{filename}.{inspection.metadata.file_type}"

        multiform_body: MultipartEncoder = self._construct_multiform_request(
            filename=filename_with_ending, inspection=inspection, metadata=metadata
        )

        self._ingest(
            inspection=inspection,
            multiform_body=multiform_body,
            request_url=request_url,
            token=token,
        )

    def _ingest(self, inspection, multiform_body, request_url, token):
        try:
            self.request_handler.post(
                url=request_url,
                params={"DataType": "still"},
                data=multiform_body,
                headers={
                    "Authorization": f"Bearer {token}",
                    "Content-Type": multiform_body.content_type,
                },
            )
        except (RequestException, HTTPError) as e:
            self.logger.warning(
                f"Failed to upload inspection: {inspection.id} to SLIMM due to a "
                f"request exception"
            )
            raise StorageException from e

    @staticmethod
    def _construct_multiform_request(filename, inspection, metadata):
        array_of_orientation = (
            inspection.metadata.time_indexed_pose.pose.orientation.to_quat_array().tolist()
        )
        multiform_body: MultipartEncoder = MultipartEncoder(
            fields={
                "SchemaMetadata.Mission.MissionId": metadata.mission_id,
                "SchemaMetadata.Mission.StartDate": metadata.mission_date.isoformat(),
                "SchemaMetadata.Mission.EndDate": metadata.mission_date.isoformat(),
                "SchemaMetadata.Geodetic.CoordinateReferenceSystemCode": metadata.coordinate_reference_system,  # noqa: E501
                "SchemaMetadata.Geodetic.VerticalCoordinateReferenceSystemCode": metadata.vertical_reference_system,  # noqa: E501
                "SchemaMetadata.Geodetic.OrientationReferenceSystem": metadata.media_orientation_reference_system,  # noqa: E501
                "SchemaMetadata.SensorCarrier.Id": metadata.robot_id,
                "SchemaMetadata.InternalClassification": metadata.data_classification,
                "SchemaMetadata.PlantFacilitySAPCode": metadata.plant_code,
                "SchemaMetadata.Mission.Client": "Equinor",
                "SchemaMetadata.IsoCountryCode": "NO",
                "AttachedFileMetadata.X": str(
                    inspection.metadata.time_indexed_pose.pose.position.x
                ),
                "AttachedFileMetadata.Y": str(
                    inspection.metadata.time_indexed_pose.pose.position.y
                ),
                "AttachedFileMetadata.Z": str(
                    inspection.metadata.time_indexed_pose.pose.position.z
                ),
                "AttachedFileMetadata.CameraOrientation[0]": str(
                    array_of_orientation[0]
                ),
                "AttachedFileMetadata.CameraOrientation[1]": str(
                    array_of_orientation[1]
                ),
                "AttachedFileMetadata.CameraOrientation[2]": str(
                    array_of_orientation[2]
                ),
                "AttachedFileMetadata.CameraOrientation[3]": str(
                    array_of_orientation[3]
                ),
                "AttachedFileMetadata.FunctionalLocation": inspection.metadata.tag_id  # noqa: E501
                if inspection.metadata.tag_id
                else "NA",
                "AttachedFileMetadata.Timestamp": inspection.metadata.start_time.isoformat(),  # noqa: E501
                "AttachedFile": (filename, inspection.data),
            }
        )
        return multiform_body



from dataclasses import dataclass
from datetime import datetime

from alitra import Pose


@dataclass
class TelemetryPosePayload:
    pose: Pose
    robot_id: str
    timestamp: datetime


@dataclass
class TelemetryBatteryPayload:
    battery_level: float
    robot_id: str
    timestamp: datetime



import logging
from typing import Optional

from alitra import Frame, Position
from azure.identity import DefaultAzureCredential
from injector import inject
from requests import Response

from isar.config.settings import settings
from isar.services.auth.azure_credentials import AzureCredentials
from isar.services.service_connections.request_handler import RequestHandler


class StidService:
    @inject
    def __init__(self, request_handler: RequestHandler):
        self.request_handler: RequestHandler = request_handler
        self.credentials: DefaultAzureCredential = (
            AzureCredentials.get_azure_credentials()
        )
        self.logger = logging.getLogger("api")

    def tag_position(self, tag: str) -> Optional[Position]:
        client_id: str = settings.STID_CLIENT_ID
        scope: str = settings.STID_APP_SCOPE
        request_scope: str = f"{client_id}/{scope}"

        token: str = self.credentials.get_token(request_scope).token

        stid_url: str = settings.STID_API_URL
        plant_name: str = settings.STID_PLANT_NAME
        request_url: str = f"{stid_url}/{plant_name}/tag"

        response: Response = self.request_handler.get(
            url=request_url,
            params={"tagNo": tag},
            headers={"Authorization": f"Bearer {token}"},
        )
        tag_metadata: dict = response.json()

        x_coord: float = tag_metadata["xCoordinate"] / 1000
        y_coord: float = tag_metadata["yCoordinate"] / 1000
        z_coord: float = tag_metadata["zCoordinate"] / 1000

        return Position(x=x_coord, y=y_coord, z=z_coord, frame=Frame("asset"))



from .idle import Idle
from .initialize import Initialize
from .initiate_step import InitiateStep
from .monitor import Monitor
from .off import Off
from .paused import Paused
from .stop_step import StopStep



import logging
import time
from typing import Callable, TYPE_CHECKING

from transitions import State

from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions import RobotException

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class StopStep(State):
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="stop_step", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.stop_thread = None
        self._count_number_retries: int = 0

    def start(self):
        self.state_machine.update_state()
        self._run()

    def stop(self):
        if self.stop_thread:
            self.stop_thread.wait_for_thread()
        self.stop_thread = None
        self._count_number_retries = 0

    def _run(self):
        transition: Callable
        while True:
            if not self.stop_thread:
                self.stop_thread = ThreadedRequest(self.state_machine.robot.stop)
                self.stop_thread.start_thread()

            if self.state_machine.should_stop_mission():
                self.state_machine.stopped = True

            try:
                self.stop_thread.get_output()
            except ThreadedRequestNotFinishedError:
                if self.handle_stop_fail(
                    retry_limit=self.state_machine.stop_robot_attempts_limit
                ):
                    transition = self.state_machine.mission_stopped
                    break
                continue

            except RobotException:
                self.logger.warning("Failed to stop robot. Retrying.")
                self.stop_thread = None
                continue
            if self.state_machine.stopped:
                transition = self.state_machine.mission_stopped
            else:
                transition = self.state_machine.mission_paused
            break

        transition()

    def handle_stop_fail(self, retry_limit: int) -> bool:
        self._count_number_retries += 1
        if self._count_number_retries > retry_limit:
            self.logger.warning(
                "Could not communicate request: Reached limit for stop attemps. Cancelled mission and transitioned to idle."
            )
            return True
        time.sleep(self.state_machine.sleep_time)
        return False



from datetime import datetime
from queue import Queue
from threading import Thread
from typing import List, Sequence

from alitra import Frame, Orientation, Pose, Position

from robot_interface.models.initialize import InitializeParams
from robot_interface.models.inspection.inspection import (
    Image,
    ImageMetadata,
    Inspection,
    TimeIndexedPose,
)
from robot_interface.models.mission import InspectionStep, Step, StepStatus
from robot_interface.robot_interface import RobotInterface


class MockRobot(RobotInterface):
    def __init__(
        self,
        initiate_step: bool = True,
        step_status: StepStatus = StepStatus.Successful,
        stop: bool = True,
        pose: Pose = Pose(
            position=Position(x=0, y=0, z=0, frame=Frame("robot")),
            orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame("robot")),
            frame=Frame("robot"),
        ),
    ):
        self.initiate_step_return_value: bool = initiate_step
        self.step_status_return_value: StepStatus = step_status
        self.stop_return_value: bool = stop
        self.robot_pose_return_value: Pose = pose

    def initiate_step(self, step: Step) -> None:
        return

    def step_status(self) -> StepStatus:
        return self.step_status_return_value

    def stop(self) -> None:
        return

    def get_inspections(self, step: InspectionStep) -> Sequence[Inspection]:
        image: Image = Image(mock_image_metadata())
        image.data = b"Some binary image data"
        return [image]

    def initialize(self, params: InitializeParams) -> None:
        return

    def get_telemetry_publishers(self, queue: Queue, robot_id: str) -> List[Thread]:
        return []


def mock_image_metadata() -> ImageMetadata:
    return ImageMetadata(
        datetime.now(),
        TimeIndexedPose(
            Pose(
                Position(0, 0, 0, Frame("robot")),
                Orientation(0, 0, 0, 1, Frame("robot")),
                Frame("robot"),
            ),
            datetime.now(),
        ),
        file_type="jpg",
    )



import logging
from copy import deepcopy
from typing import Any, Optional

from alitra import Pose
from injector import inject
from isar.apis.models.models import (
    ControlMissionResponse,
    StartMissionResponse,
)

from isar.config.settings import settings
from isar.models.communication.message import StartMissionMessage
from isar.models.communication.queues import QueueIO, Queues, QueueTimeoutError
from isar.models.mission.mission import Mission
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.state_machine.states_enum import States


class SchedulingUtilities:
    """
    Contains utility functions for scheduling missions from the API. The class handles
    required thread communication through queues to the state machine.
    """

    @inject
    def __init__(self, queues: Queues, queue_timeout: int = settings.QUEUE_TIMEOUT):
        self.queues = queues
        self.queue_timeout: int = queue_timeout
        self.logger = logging.getLogger("api")

    def get_state(self) -> States:
        return self.queues.state.check()

    def start_mission(self, mission: Mission, initial_pose: Optional[Pose]) -> None:
        self._send_command(
            StartMissionMessage(mission=deepcopy(mission), initial_pose=initial_pose),
            self.queues.start_mission,
        )

    def pause_mission(self) -> ControlMissionResponse:
        return self._send_command(True, self.queues.pause_mission)

    def resume_mission(self) -> ControlMissionResponse:
        return self._send_command(True, self.queues.resume_mission)

    def stop_mission(self) -> ControlMissionResponse:
        return self._send_command(True, self.queues.stop_mission)

    def _send_command(self, input: Any, queueio: QueueIO) -> Any:
        queueio.input.put(input)
        try:
            return QueueUtilities.check_queue(
                queueio.output,
                self.queue_timeout,
            )
        except QueueTimeoutError as e:
            QueueUtilities.clear_queue(queueio.input)
            self.logger.error("Timeout while communicating with state machine")
            raise e



from enum import Enum


class StepStatus(str, Enum):

    NotStarted: str = "not_started"
    Successful: str = "successful"
    InProgress: str = "in_progress"
    Failed: str = "failed"
    Cancelled: str = "cancelled"



from alitra import Frame, Position

from robot_interface.models.mission import DriveToPose, TakeImage
from tests.mocks.pose import MockPose


class MockStep:
    drive_to = DriveToPose(pose=MockPose.default_pose)
    take_image_in_coordinate_direction = TakeImage(
        target=Position(x=1, y=1, z=1, frame=Frame("robot"))
    )



from alitra import Frame, Orientation, Pose, Position


class MockPose:
    default_pose = Pose(
        position=Position(x=0, y=0, z=0, frame=Frame("robot")),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame("robot")),
        frame=Frame("robot"),
    )



import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from queue import Empty, Queue
from typing import List

from isar.config.settings import settings
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.storage.storage_interface import StorageException, StorageInterface
from robot_interface.models.inspection.inspection import Inspection


@dataclass
class UploaderQueueItem:
    inspection: Inspection
    mission_metadata: MissionMetadata
    storage_handler: StorageInterface
    _retry_count: int
    _next_retry_time: datetime = datetime.utcnow()

    def increment_retry(self, max_wait_time: int):
        self._retry_count += 1
        seconds_until_retry: int = min(2**self._retry_count, max_wait_time)
        self._next_retry_time = datetime.utcnow() + timedelta(
            seconds=seconds_until_retry
        )

    def get_retry_count(self) -> int:
        return self._retry_count

    def is_ready_for_upload(self) -> bool:
        return datetime.utcnow() >= self._next_retry_time

    def seconds_until_retry(self) -> int:
        return max(0, int((self._next_retry_time - datetime.utcnow()).total_seconds()))


class Uploader:
    def __init__(
        self,
        upload_queue: Queue,
        storage_handlers: List[StorageInterface],
        max_wait_time: int = settings.UPLOAD_FAILURE_MAX_WAIT,
        max_retry_attempts: int = settings.UPLOAD_FAILURE_ATTEMPTS_LIMIT,
    ) -> None:
        """Initializes the uploader.

        Parameters
        ----------
        upload_queue : Queue
            Queue used for cross-thread communication.
        storage_handlers : List[StorageInterface]
            List of handlers for different upload options
        max_wait_time : float
            The maximum wait time between two retries (exponential backoff)
        max_retry_attempts : int
            Maximum attempts to retry an upload when it fails
        """
        self.upload_queue: Queue = upload_queue
        self.storage_handlers: List[StorageInterface] = storage_handlers
        self.max_wait_time = max_wait_time
        self.max_retry_attempts = max_retry_attempts
        self._internal_upload_queue: List[UploaderQueueItem] = []

        self.logger = logging.getLogger("uploader")

    def run(self) -> None:
        self.logger.info("Started uploader")
        while True:
            inspection: Inspection
            mission_metadata: MissionMetadata
            try:
                if self._internal_upload_queue:
                    self._process_upload_queue()

                inspection, mission_metadata = self.upload_queue.get(timeout=1)

                # If new item from thread queue, add one per handler to internal queue:
                for storage_handler in self.storage_handlers:
                    new_item: UploaderQueueItem = UploaderQueueItem(
                        inspection, mission_metadata, storage_handler, _retry_count=-1
                    )
                    self._internal_upload_queue.append(new_item)
            except Empty:
                continue

    def _upload(self, upload_item: UploaderQueueItem):
        try:
            upload_item.storage_handler.store(
                inspection=upload_item.inspection, metadata=upload_item.mission_metadata
            )
            self.logger.info(
                f"Storage handler: {type(upload_item.storage_handler).__name__} "
                f"uploaded inspection {str(upload_item.inspection.id)[:8]}"
            )
            self._internal_upload_queue.remove(upload_item)
        except StorageException:
            if upload_item.get_retry_count() < self.max_retry_attempts:
                upload_item.increment_retry(self.max_wait_time)
                self.logger.warning(
                    f"Storage handler: {type(upload_item.storage_handler).__name__} "
                    f"failed to upload inspection: "
                    f"{str(upload_item.inspection.id)[:8]}. "
                    f"Retrying in {upload_item.seconds_until_retry()}s."
                )
            else:
                self._internal_upload_queue.remove(upload_item)
                self.logger.error(
                    f"Storage handler: {type(upload_item.storage_handler).__name__} "
                    f"exceeded max retries to upload inspection: "
                    f"{str(upload_item.inspection.id)[:8]}. Aborting upload."
                )

    def _process_upload_queue(self):
        ready_items: List[UploaderQueueItem] = [
            x for x in self._internal_upload_queue if x.is_ready_for_upload()
        ]
        for item in ready_items:
            self._upload(item)



from alitra import Frame, Orientation, Pose, Position

# Euler heading to quaternion
# East = 0 deg = (x=0, y=0, z=0, w=1)
# North = 90 deg = (x=0, y=0, z=0.7071, w=0.7071)
# West = 180 deg = (x=0, y=0, z=1, w=0)
# South = 270 deg = (x=0, y=0, z=-0.7071, w=0.7071)

robot_frame = Frame("robot")
asset_frame = Frame("asset")
predefined_poses = {
    #
    # K-lab VG2
    #
    # View from VG2 towards crane
    "20-PT-5583": Pose(
        position=Position(x=20257.979, y=5310.630, z=14.365, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # Open area picture of fuse box
    "EJE-342-0231": Pose(
        position=Position(x=20252.559, y=5308.437, z=14.375, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # Big valve straight behind the stairs. East heading
    "331-LD-525": Pose(
        position=Position(x=20254.540, y=5311.651, z=14.321, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # Small valve close to utility station. East heading
    "344-LD-1024": Pose(
        position=Position(x=20260.593, y=5312.311, z=14.418, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    #
    # K-lab inlet area
    #
    "334-LD-0225": Pose(
        position=Position(x=1.63, y=1.724, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.5383686, w=0.8427095, frame=robot_frame),
        frame=robot_frame,
    ),
    "314-LD-1001": Pose(
        position=Position(x=24.853, y=23.761, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.998122, w=0.0612579, frame=robot_frame),
        frame=robot_frame,
    ),
    "346-LD-1073": Pose(
        position=Position(x=25.005, y=23.607, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.9995287, w=-0.0306988, frame=robot_frame),
        frame=robot_frame,
    ),
    "314-PI-001": Pose(
        position=Position(x=25.041, y=23.682, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.8907533, w=0.4544871, frame=robot_frame),
        frame=robot_frame,
    ),
    "300-LD-0025": Pose(
        position=Position(x=21.279, y=17.392, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.2006291, w=0.9796673, frame=robot_frame),
        frame=robot_frame,
    ),
    "344-LD-1001": Pose(
        position=Position(x=24.853, y=23.761, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.998122, w=0.0612579, frame=robot_frame),
        frame=robot_frame,
    ),
    "15-LD-0059": Pose(
        position=Position(x=27.297, y=22.686, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.9631559, w=0.2689439, frame=robot_frame),
        frame=robot_frame,
    ),
    "345-LD-1004": Pose(
        position=Position(x=20.293, y=20.982, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.5256679, w=0.8506899, frame=robot_frame),
        frame=robot_frame,
    ),
    # start of narrow passage
    "345-LD-003": Pose(
        position=Position(x=20.994, y=10.3, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.8963647, w=0.4433175, frame=robot_frame),
        frame=robot_frame,
    ),
    # end of narrow passage
    "345-LD-004": Pose(
        position=Position(x=16.609, y=15.444, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0.999986, w=0.0052963, frame=robot_frame),
        frame=robot_frame,
    ),
    #
    # Compressor area K-lab
    #
    "355-LD-1003": Pose(
        position=Position(x=20248.440, y=5247.118, z=14.450, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.054, w=0.998, frame=asset_frame),
        frame=asset_frame,
    ),
    "313-LD-1111": Pose(
        position=Position(x=20249.830, y=5246.737, z=14.450, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.999, w=-0.041, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1104 and 313-PA-101A has the same pose
    "313-LD-1104": Pose(
        position=Position(x=20250.720, y=5252.582, z=14.450, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.356, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1104 and 313-PA-101A has the same pose
    "313-PA-101A": Pose(
        position=Position(x=20252.860, y=5252.368, z=14.450, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    #
    #   Second Floor K-Lab Compressor area
    #
    # 300-LD-0066 and 300-XCV-003 has the same pose
    # Robot Orientation: 90 deg
    "300-LD-0066": Pose(
        position=Position(x=20259.220, y=5241.849, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 300-LD-0066 and 300-XCV-003 has the same pose
    # Robot Orientation: 90 deg
    "300-XCV-003": Pose(
        position=Position(x=20259.220, y=5241.849, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 0 deg
    "313-LD-1248": Pose(
        position=Position(x=20256.540, y=5243.902, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 0 deg
    "313-FG-136": Pose(
        position=Position(x=20256.470, y=5243.299, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1177 and 313-LD-1135 has the same pose
    # Robot Orientation: 270 deg
    "313-LD-1177": Pose(
        position=Position(x=20254.740, y=5242.001, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1177 and 313-LD-1135 has the same pose
    # Robot Orientation: 270 deg
    "313-LD-1135": Pose(
        position=Position(x=20254.740, y=5242.001, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 270 deg
    "313-LD-1133": Pose(
        position=Position(x=20252.960, y=5241.896, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 270 deg
    "313-FG-135": Pose(
        position=Position(x=20252.100, y=5241.975, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1243 and 313-LD-1242 has the same pose
    # Robot Orientation: 180 deg
    "313-LD-1243": Pose(
        position=Position(x=20256.640, y=5243.246, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1243 and 313-LD-1242 has the same pose
    # Robot Orientation: 180 deg
    "313-LD-1242": Pose(
        position=Position(x=20256.640, y=5243.246, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 270 deg
    "313-LD-1037": Pose(
        position=Position(x=20254.650, y=5242.229, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 270 deg
    "313-LD-1050": Pose(
        position=Position(x=20252.560, y=5241.928, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 270 deg
    "313-TT-1102C": Pose(
        position=Position(x=20252.250, y=5242.033, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1241 and 313-LD-1240 has the same pose
    # Robot Orientation: 270 deg
    "313-LD-1241": Pose(
        position=Position(x=20252.210, y=5241.930, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1241 and 313-LD-1240 has the same pose
    # Robot Orientation: 270 deg
    "313-LD-1240": Pose(
        position=Position(x=20252.210, y=5241.930, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 270 deg
    "313-LD-1056": Pose(
        position=Position(x=20256.470, y=5241.838, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 270 deg
    "313-LD-1060": Pose(
        position=Position(x=20253.920, y=5246.332, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # Robot Orientation: 90 deg
    "313-PT-1012": Pose(
        position=Position(x=20252.940, y=5246.394, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-PT-1061B and 313-LD-1257 has the same pose
    # Robot Orientation: 90 deg
    "313-PT-1016B": Pose(
        position=Position(x=20250.430, y=5246.008, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-PT-1061B and 313-LD-1257 has the same pose
    # Robot Orientation: 90 deg
    "313-LD-1257": Pose(
        position=Position(x=20250.430, y=5246.008, z=16.880, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 355-LD-1079 Robot orientation 90 degrees
    "355-LD-1079": Pose(
        position=Position(x=20252.630, y=5253.167, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 331-EV-1009 Robot orientation 90 degrees
    "331-EV-1009": Pose(
        position=Position(x=20255.710, y=5252.991, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1094 Robot orientation 90 degrees
    "313-LD-1094": Pose(
        position=Position(x=20255.710, y=5252.980, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1095 Robot orientation 135 degrees
    "313-LD-1095": Pose(
        position=Position(x=20256.860, y=5252.572, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1213 Robot orientation 180 degrees
    "313-LD-1213": Pose(
        position=Position(x=20257.070, y=5250.289, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-LD-1214 Robot orientation 180 degrees
    "313-LD-1214": Pose(
        position=Position(x=20257.070, y=5250.289, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 300-LD-1001 Robot orientation 235 degrees. Part above in 4th floor
    "300-LD-1001": Pose(
        position=Position(x=20257.000, y=5248.983, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 300-LD-1002 Robot orientation 235 degrees. Part above in 4th floor
    "300-LD-1002": Pose(
        position=Position(x=20257.000, y=5248.983, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-JA-001 Robot orientation 270 degrees
    "313-JA-001": Pose(
        position=Position(x=20255.810, y=5248.977, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-HH-001 Robot orientation 0 degrees. Should have photo from other side?
    # Implementation of handling several photos for same part?
    "313-HH-001": Pose(
        position=Position(x=20250.810, y=5249.645, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # EJE-342-1004.08 Robot orientation 0 degrees
    "EJE-342-1004.08": Pose(
        position=Position(x=20251.050, y=5251.786, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # EJE-342-0226.10 Robot orientation 90 degrees
    "EJE-342-0226.10": Pose(
        position=Position(x=20252.070, y=5253.136, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # EJE-342-1004.03 Robot orientation 90 degrees
    "EJE-342-1004.03": Pose(
        position=Position(x=20252.630, y=5253.136, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # 313-KG-001 Robot orientation 90 degrees. Pose on stairs
    "313-KG-001": Pose(
        position=Position(x=20253.900, y=5253.020, z=18.735, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # JBZS-313-021 Robot orientation 270 degrees
    "JBZS-313-021": Pose(
        position=Position(x=20250.000, y=5253.155, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # JBZS-313-024 Robot orientation 315 degrees
    "JBZS-313-024": Pose(
        position=Position(x=20250.160, y=5249.555, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    # JBES-313-003 Robot orientation 315 degrees
    "JBES-313-003": Pose(
        position=Position(x=20250.160, y=5249.555, z=18.540, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.935, w=0.357, frame=asset_frame),
        frame=asset_frame,
    ),
    #
    # AP520 & AP530
    #
    # A-VB20-0292 Car seal valve south heading
    "A-VB20-0292": Pose(
        position=Position(x=309.978, y=108.173, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB20-0656 Car seal valve south heading
    "A-VB20-0656": Pose(
        position=Position(x=309.931, y=103.610, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0111 Car seal valve east heading
    "A-VB23-0111": Pose(
        position=Position(x=319.502, y=90.022, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0118 Car seal valve west heading
    "A-VB23-0118": Pose(
        position=Position(x=319.865, y=112.225, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=1, w=0, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB50-0119 Car seal valve south heading
    "A-VB50-0119": Pose(
        position=Position(x=312.504, y=102.436, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB50-0150 Car seal valve east heading
    "A-VB50-0150": Pose(
        position=Position(x=316.401, y=110.011, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB64-0095 Car seal valve south heading
    "A-VB64-0095": Pose(
        position=Position(x=310.027, y=105.676, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB64-0096 Car seal valve south heading
    "A-VB64-0096": Pose(
        position=Position(x=310.120, y=104.511, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB64-0100 Car seal valve south heading
    "A-VB64-0100": Pose(
        position=Position(x=310.041, y=100.015, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB64-0101 Car seal valve south heading
    "A-VB64-0101": Pose(
        position=Position(x=310.041, y=100.015, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0147 Car seal valve east heading
    "A-VB23-0147": Pose(
        position=Position(x=333.650, y=112.330, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0392 Car seal valve east heading
    "A-VB23-0392": Pose(
        position=Position(x=334.744, y=90.333, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0539 Car seal valve east heading
    "A-VB23-0539": Pose(
        position=Position(x=335.099, y=112.361, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0398 Car seal valve east heading
    "A-VB23-0398": Pose(
        position=Position(x=343.198, y=87.717, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0545 Car seal valve east heading
    "A-VB23-0545": Pose(
        position=Position(x=343.959, y=112.298, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VF50-0153 Car seal valve east heading
    "A-VF50-0153": Pose(
        position=Position(x=338.766, y=101.248, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VF50-0154 Car seal valve east heading
    "A-VF50-0154": Pose(
        position=Position(x=337.611, y=109.592, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB23-0183 Car seal valve east heading
    "A-VB23-0183": Pose(
        position=Position(x=341.412, y=83.433, z=536.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    #
    # AP430
    #
    # A-VB20-0039 Car seal valve East
    "A-VB20-0039": Pose(
        position=Position(x=328.343, y=83.986, z=531.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VG23-0104 Car seal valve South
    "A-VG23-0104": Pose(
        position=Position(x=323.073, y=101.064, z=531.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.707, w=0.707, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VG40-0367 Car seal valve North
    "A-VG40-0367": Pose(
        position=Position(x=332.231, y=92.935, z=531.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.707, w=0.707, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VM23-0399 Car seal valve South
    "A-VM23-0399": Pose(
        position=Position(x=344.574, y=88.770, z=531.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.707, w=0.707, frame=asset_frame),
        frame=asset_frame,
    ),
    # A-VB24-0032 Car seal valve South
    "A-VB24-0032": Pose(
        position=Position(x=341.830, y=90.141, z=531.850, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.707, w=0.707, frame=asset_frame),
        frame=asset_frame,
    ),
    # Home on JS weather deck
    # A-72SP123 Fire extinguisher east heading
    "A-72SP123": Pose(
        position=Position(x=13.453, y=10.317, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=robot_frame),
        frame=robot_frame,
    ),
    # Home on JS intermediate deck
    #  Fire extinguisher east heading
    "A-72SP102": Pose(
        position=Position(x=17, y=10.9, z=0, frame=robot_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=robot_frame),
        frame=robot_frame,
    ),
    #
    # AQ200
    #
    "VG29-4269": Pose(
        position=Position(x=155.5, y=289.8, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=1, w=0, frame=asset_frame),
        frame=asset_frame,
    ),
    "29PT0183": Pose(
        position=Position(x=152.5, y=289.9, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=1, w=0, frame=asset_frame),
        frame=asset_frame,
    ),
    "VB64-4037": Pose(
        position=Position(x=149.25, y=292.16, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=1, w=0, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG53-4024": Pose(
        position=Position(x=151.5, y=286, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VF29-4270": Pose(
        position=Position(x=153.35, y=285.56, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4352": Pose(
        position=Position(x=155.2, y=285.63, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4249": Pose(
        position=Position(x=154.24, y=282.04, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=1, w=0, frame=asset_frame),
        frame=asset_frame,
    ),
    "29LT0112": Pose(
        position=Position(x=150.88, y=279.84, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=1, w=0, frame=asset_frame),
        frame=asset_frame,
    ),
    "29XU0188": Pose(
        position=Position(x=157.46, y=277.14, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=-0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VB71-4216": Pose(
        position=Position(x=160.75, y=276.86, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4337": Pose(
        position=Position(x=160.1, y=278.2, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4335": Pose(
        position=Position(x=160.07, y=279.07, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4333": Pose(
        position=Position(x=160.01, y=279.78, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4331": Pose(
        position=Position(x=160.1, y=280.65, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4329": Pose(
        position=Position(x=160.12, y=281.56, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4327": Pose(
        position=Position(x=160.08, y=282.39, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG29-4325": Pose(
        position=Position(x=160.08, y=282.54, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0.7071, w=0.7071, frame=asset_frame),
        frame=asset_frame,
    ),
    "29PT0214": Pose(
        position=Position(x=161.81, y=282.98, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
    "72SP211": Pose(
        position=Position(x=147.07, y=280.34, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=1, w=0, frame=asset_frame),
        frame=asset_frame,
    ),
    "VG63-4001 ": Pose(
        position=Position(x=148.19, y=287.14, z=40.8, frame=asset_frame),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=asset_frame),
        frame=asset_frame,
    ),
}



from enum import Enum


class MissionStatus(str, Enum):
    NotStarted: str = "not_started"
    Started: str = "started"
    InProgress: str = "in_progress"
    Failed: str = "failed"
    Cancelled: str = "cancelled"
    Completed: str = "completed"
    Paused: str = "paused"


class TaskStatus(str, Enum):
    NotStarted: str = "not_started"
    InProgress: str = "in_progress"
    PartiallySuccessful: str = "partially_successful"
    Failed: str = "failed"
    Cancelled: str = "cancelled"
    Successful: str = "successful"
    Paused: str = "paused"



import logging
from http import HTTPStatus
from queue import Empty
from typing import List, Optional, Union

from alitra import Pose
from fastapi import Body, Path, Response
from injector import inject
from requests import HTTPError

from isar.apis.models import InputPose, StartMissionResponse
from isar.apis.models.models import ControlMissionResponse
from isar.apis.models.start_mission_definition import (
    StartMissionDefinition,
    to_isar_mission,
)
from isar.config.settings import robot_settings, settings
from isar.mission_planner.mission_planner_interface import (
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.mission_planner.mission_validator import is_robot_capable_of_mission
from isar.models.communication.queues import QueueTimeoutError
from isar.models.mission import Mission, Task
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from robot_interface.models.mission import DriveToPose


class SchedulingController:
    @inject
    def __init__(
        self,
        mission_planner: MissionPlannerInterface,
        scheduling_utilities: SchedulingUtilities,
        queue_timeout: int = settings.QUEUE_TIMEOUT,
    ):
        self.logger = logging.getLogger("api")
        self.mission_planner: MissionPlannerInterface = mission_planner
        self.scheduling_utilities: SchedulingUtilities = scheduling_utilities
        self.queue_timeout: int = queue_timeout

    def start_mission_by_id(
        self,
        response: Response,
        mission_id: int = Path(
            ...,
            alias="id",
            title="Mission ID",
            description="ID-number for predefined mission",
        ),
        initial_pose: Optional[InputPose] = Body(
            default=None,
            description="The starting point of the mission. Used for initial "
            "localization of robot",
            embed=True,
        ),
        return_pose: Optional[InputPose] = Body(
            default=None,
            description="End pose of the mission. The robot return to the specified "
            "pose after finishing all inspections",
            embed=True,
        ),
    ):
        self.logger.info(f"Received request to start mission with id {mission_id}")
        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            error_message: str = (
                "Internal Server Error - Current state of state machine unknown"
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return error_message

        if state in [
            States.Initialize,
            States.InitiateStep,
            States.StopStep,
            States.Monitor,
            States.Paused,
        ]:
            error_message = "Conflict - Mission already in progress"
            self.logger.warning(error_message)
            response.status_code = HTTPStatus.CONFLICT.value
            return error_message

        try:
            mission: Mission = self.mission_planner.get_mission(mission_id)
        except HTTPError as e:
            self.logger.error(e)
            response.status_code = e.response.status_code
            raise
        except MissionPlannerError as e:
            self.logger.error(e)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return

        robot_capable: bool
        missing_functions: List[str]
        (robot_capable, missing_functions) = is_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )
        if not robot_capable:
            error_message = (
                "Bad Request - Robot is not capable of performing mission. Missing "
                "functionalities: " + str(missing_functions)
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.BAD_REQUEST.value
            return error_message

        initial_pose_alitra: Optional[Pose]
        if initial_pose:
            initial_pose_alitra = initial_pose.to_alitra_pose()
        else:
            initial_pose_alitra = None

        self.logger.info(f"Starting mission: {mission.id}")

        if return_pose:
            pose: Pose = return_pose.to_alitra_pose()
            step: DriveToPose = DriveToPose(pose=pose)
            mission.tasks.append(Task(steps=[step]))

        try:
            self.scheduling_utilities.start_mission(
                mission=mission, initial_pose=initial_pose_alitra
            )
            self.logger.info("OK - Mission successfully started")
        except QueueTimeoutError:
            error_message = "Timeout - Failed to start mission"
            self.logger.error(error_message)
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            return error_message
        return StartMissionResponse(**mission.api_response_dict())

    def start_mission(
        self,
        response: Response,
        mission_definition: StartMissionDefinition = Body(
            default=None,
            embed=True,
            title="Mission Definition",
            description="Description of the mission in json format",
        ),
        initial_pose: Optional[InputPose] = Body(
            default=None,
            description="The starting point of the mission. Used for initial "
            "localization of robot",
            embed=True,
        ),
        return_pose: Optional[InputPose] = Body(
            default=None,
            description="End pose of the mission. The robot return to the specified "
            "pose after finishing all inspections",
            embed=True,
        ),
    ):
        self.logger.info("Received request to start new mission")

        if not mission_definition:
            error_message: str = (
                "Unprocessable entity - 'mission_definition' empty or invalid"
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.UNPROCESSABLE_ENTITY.value
            return error_message

        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            error_message = (
                "Internal Server Error - Current state of state machine unknown"
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return error_message

        if state in [
            States.Initialize,
            States.InitiateStep,
            States.StopStep,
            States.Monitor,
            States.Paused,
        ]:
            error_message = "Conflict - Mission already in progress"
            self.logger.warning(error_message)
            response.status_code = HTTPStatus.CONFLICT.value
            return error_message

        try:
            mission: Mission = to_isar_mission(mission_definition)
        except MissionPlannerError as e:
            error_message = f"Bad Request - Cannot create ISAR mission: {e}"
            self.logger.warning(error_message)
            response.status_code = HTTPStatus.BAD_REQUEST.value
            return error_message

        robot_capable: bool
        missing_functions: List[str]
        (robot_capable, missing_functions) = is_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )
        if not robot_capable:
            error_message = (
                "Bad Request - Robot is not capable of performing mission. Missing "
                "functionalities: " + str(missing_functions)
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.BAD_REQUEST.value
            return error_message

        initial_pose_alitra: Optional[Pose]
        if initial_pose:
            initial_pose_alitra = initial_pose.to_alitra_pose()
        else:
            initial_pose_alitra = None

        self.logger.info(f"Starting mission: {mission.id}")

        if return_pose:
            pose: Pose = return_pose.to_alitra_pose()
            step: DriveToPose = DriveToPose(pose=pose)
            mission.tasks.append(Task(steps=[step]))
        try:
            self.scheduling_utilities.start_mission(
                mission=mission, initial_pose=initial_pose_alitra
            )
            self.logger.info("OK - Mission successfully started")
        except QueueTimeoutError:
            error_message = "Timeout - Failed to start mission"
            self.logger.error(error_message)
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            return error_message
        return StartMissionResponse(**mission.api_response_dict())

    def pause_mission(self, response: Response) -> Union[ControlMissionResponse, str]:
        self.logger.info("Received request to pause current mission")

        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            errorMsg = "Internal Server Error - Current state of state machine unknown"
            self.logger.error(errorMsg)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return errorMsg

        if state in [States.Idle, States.StopStep, States.Paused, States.Initialize]:
            errorMsg = "Conflict - Pause command received in invalid state"
            self.logger.warning(errorMsg)
            response.status_code = HTTPStatus.CONFLICT.value
            return errorMsg

        try:
            pause_mission_response: ControlMissionResponse = (
                self.scheduling_utilities.pause_mission()
            )
            self.logger.info("OK - Mission successfully paused")
        except QueueTimeoutError:
            errorMsg = "Timeout - Failed to pause mission"
            self.logger.error(errorMsg)
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            return errorMsg
        return pause_mission_response

    def resume_mission(self, response: Response) -> Union[ControlMissionResponse, str]:
        self.logger.info("Received request to resume current mission")
        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            error_message = (
                "Internal Server Error - Current state of state machine unknown"
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return error_message

        if state in [
            States.Idle,
            States.InitiateStep,
            States.Monitor,
            States.StopStep,
            States.Initialize,
        ]:
            error_message = "Conflict - Resume command received in invalid state"
            self.logger.warning(error_message)
            response.status_code = HTTPStatus.CONFLICT.value
            return error_message

        try:
            resume_mission_response: ControlMissionResponse = (
                self.scheduling_utilities.resume_mission()
            )
            self.logger.info("OK - Mission successfully resumed")
        except QueueTimeoutError:
            error_message = "Timeout - Failed to resume mission"
            self.logger.error(error_message)
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            return error_message
        return resume_mission_response

    def stop_mission(self, response: Response) -> Union[ControlMissionResponse, str]:
        self.logger.info("Received request to stop current mission")

        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            error_message = (
                "Internal Server Error - Current state of state machine unknown"
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return error_message

        if state in [States.Idle, States.Initialize]:
            error_message = "Conflict - Stop command received in invalid state"
            self.logger.warning(error_message)
            response.status_code = HTTPStatus.CONFLICT.value
            return error_message

        try:
            stop_mission_response: ControlMissionResponse = (
                self.scheduling_utilities.stop_mission()
            )
            self.logger.info("OK - Mission successfully stopped")
        except QueueTimeoutError:
            error_message = "Timeout - Failed to stop mission"
            self.logger.error(error_message)
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            return error_message
        return stop_mission_response

    def drive_to(
        self,
        response: Response,
        target_pose: InputPose = Body(
            default=None,
            title="Target Pose",
            description="The target pose for the drive_to step",
        ),
    ):
        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            error_message = (
                "Internal Server Error - Current state of state machine unknown"
            )
            self.logger.error(error_message)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return error_message

        if state in [
            States.Initialize,
            States.InitiateStep,
            States.StopStep,
            States.Monitor,
            States.Paused,
        ]:
            error_message = "Conflict - DriveTo command received in invalid state"
            self.logger.warning(error_message)
            response.status_code = HTTPStatus.CONFLICT.value
            return error_message

        pose: Pose = target_pose.to_alitra_pose()

        step: DriveToPose = DriveToPose(pose=pose)
        mission: Mission = Mission(tasks=[Task(steps=[step])])

        try:
            self.scheduling_utilities.start_mission(mission=mission, initial_pose=None)
            self.logger.info("OK - Drive to successfully started")
        except QueueTimeoutError:
            error_message = "Timeout - Failed to start DriveTo"
            self.logger.error(error_message)
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            return error_message



import json
from dataclasses import asdict, is_dataclass
from datetime import date, datetime
from typing import Iterator
from uuid import UUID

import numpy as np
from alitra import Orientation


class EnhancedJSONEncoder(json.JSONEncoder):
    """
    Custom JSONEncoder with the ability to encode dataclasses.
    """

    def default(self, o):
        if is_dataclass(o):
            return asdict(o)
        if isinstance(o, UUID):
            return str(o)
        if isinstance(o, Orientation):
            return o.__dict__
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, date):
            return o.isoformat()
        if isinstance(o, bytes):
            return "<<non-serializable: bytes>>"
        if isinstance(o, Iterator):
            return "<<non-serializable: iterator>>"
        return super().default(o)
